import os
import time

import pytest
from twisted.internet.defer import ensureDeferred

from epyqlib.tabs.files.activity_log import ActivityLog, Event

# noinspection PyUnresolvedReferences
from epyqlib.tests.utils.test_fixtures import temp_dir


@pytest.inlineCallbacks
def test_activity_log(temp_dir):
    activity_log = ActivityLog(temp_dir)

    class Listener:
        def __init__(self):
            self.type = ""

        def inc(self, event: Event):
            self.type = event.type

    listener = Listener()

    activity_log.register_listener(listener.inc)
    yield ensureDeferred(activity_log.add(Event.new_push_to_inverter("", "")))

    assert listener.type == "push-to-inverter"


@pytest.inlineCallbacks
def test_removing(temp_dir):
    activity_log = ActivityLog(temp_dir)

    class RemovingListener:
        def event(self, event: Event):
            activity_log.remove(event)

    assert len(activity_log._activity_cache) == 0

    yield ensureDeferred(activity_log.add(Event.new_push_to_inverter("", "")))

    assert len(activity_log._activity_cache) == 1

    activity_log.register_listener(RemovingListener().event)
    yield ensureDeferred(activity_log.add(Event.new_push_to_inverter("", "")))

    assert len(activity_log._activity_cache) == 1


@pytest.inlineCallbacks
def test_writing_to_file(temp_dir):
    activity_log = ActivityLog(temp_dir)
    print(f"Using: {temp_dir}")
    cache_file = os.path.join(temp_dir, "activity-cache.json")
    assert not os.path.exists(cache_file)

    yield ensureDeferred(activity_log.add(Event.new_push_to_inverter("", "")))

    assert os.path.exists(cache_file)


@pytest.inlineCallbacks
def test_writing_and_reading_from_file(temp_dir):
    activity_log = ActivityLog(temp_dir)
    print(f"Using: {temp_dir}")
    event1 = Event.new_push_to_inverter("testInvId", "testUserId")
    yield ensureDeferred(activity_log.add(event1))

    activity_log = ActivityLog(temp_dir)

    assert activity_log.has_cached_events() is False
    activity_log._read_cache_file()

    event2 = activity_log.read_oldest_event()
    assert event1.inverter_id == event2.inverter_id


@pytest.inlineCallbacks
@pytest.mark.skip("Just here for benchmarking to make sure it's not too slow")
def test_benchmark_mass_writes(temp_dir):
    activity_log = ActivityLog(temp_dir)
    print(f"Using: {temp_dir}")

    start = time.time()
    for _ in range(1000):
        event = Event.new_push_to_inverter("testInvId", "testUserId")
        yield ensureDeferred(activity_log.add(event))
    elapsed = time.time() - start
    print(f"Writing 1000 events took {elapsed:.2f}")

    time.sleep(1)

    start = time.time()
    event = Event.new_push_to_inverter("testInvId", "testUserId")
    yield ensureDeferred(activity_log.add(event))
    elapsed = time.time() - start
    print(f"Writing last event took {elapsed:.2f}")

    activity_log = ActivityLog(temp_dir)
    start = time.time()
    activity_log._read_cache_file()
    elapsed = time.time() - start
    print(f"Reading 1001 events took {elapsed:.2f}")



import pint


registry = pint.UnitRegistry()
registry.define("percent = 0.01*count = %")


def to_unitless(value, unit):
    if isinstance(value, pint.Quantity):
        return value.to(unit).magnitude

    return value



#!/usr/bin/env python3

import argparse
import collections
import epyqlib.device
import epyqlib.utils.general

try:
    import git
except ImportError as e:
    raise ImportError("Package gitpython expected but not found") from e
import json
import os
import pip
import shutil
import stat
import sys
import tempfile
import traceback
import zipfile


# TODO: CAMPid 0238493420143087667542054268097120437916848
# http://stackoverflow.com/a/21263493/228539
def del_rw(action, name, exc):
    os.chmod(name, stat.S_IWRITE)
    if os.path.isdir(name):
        os.rmdir(name)
    else:
        os.remove(name)


def collect_blobs(tree):
    blobs = tree.blobs
    for subtree in tree.trees:
        blobs.extend(collect_blobs(subtree))

    return blobs


class DeviceLoadError(Exception):
    pass


def collect(
    devices, output_directory, dry_run, groups=None, device_path=None, in_repo=True
):
    if groups is None:
        groups = []

    with tempfile.TemporaryDirectory() as checkout_dir:
        os.makedirs(output_directory, exist_ok=True)

        all_urls = set()
        all_repos = []
        all_remotes = set()
        all_devices = set()

        dirs = []

        for name, values in devices.items():
            repository = values.get("repository", None)
            print("  Handling {}".format(name))
            if repository is not None:
                dir = os.path.join(checkout_dir, name)
                dirs.append(dir)

                print("    Cloning {}".format(values["repository"]))
                repo = git.Repo.clone_from(values["repository"], dir)
                all_repos.append(repo)
                repo.git.checkout(values["branch"])
                sha = repo.head.commit.hexsha

                if values["repository"] not in all_urls:
                    all_urls.add(values["repository"])
                    all_remotes.add(repo)
                all_devices.add(
                    (values["repository"], values["branch"], values["file"])
                )
            else:
                dir = os.path.dirname(device_path)

                if in_repo:
                    repo = git.Repo(dir, search_parent_directories=True)

                    sha = "{dirty}-{sha}".format(
                        sha=repo.head.commit.hexsha,
                        dirty="local{}".format("_dirty" if repo.is_dirty() else ""),
                    )
                else:
                    sha = "not_applicable"

            device_groups = values.get("groups", [])
            if set(device_groups).isdisjoint(set(groups)):
                print("    Skipping {}".format(values["file"]))
                print(
                    "        Device groups {} not in selected groups {}".format(
                        device_groups, groups
                    )
                )
                continue

            device_path = os.path.join(dir, values["file"])
            print("    Loading {}".format(values["file"]))
            try:
                device = epyqlib.device.Device(file=device_path, only_for_files=True)
            except Exception as e:
                try:
                    raise DeviceLoadError(
                        "Unable to open '{}'".format(device_path)
                    ) from e
                except:
                    print()
                    traceback.print_exc()
                    print()
                    continue
            device_dir = os.path.dirname(device_path)
            device_file_name = os.path.basename(device_path)
            referenced_files = [device_file_name, *device.referenced_files]

            zip_file_name = name + ".epz"
            zip_path = os.path.join(output_directory, zip_file_name)
            print(
                "    {} {}".format(
                    {False: "Writing", True: "Not writing"}[dry_run], zip_file_name
                )
            )

            if not dry_run:
                epyqlib.utils.general.write_device_to_zip(
                    zip_path=zip_path,
                    checkout_dir=checkout_dir,
                    epc_dir=device_dir,
                    referenced_files=referenced_files,
                    sha=sha,
                )

        all_devices_strings = [
            "{}:{}:{}".format(url, branch, file) for url, branch, file in all_devices
        ]

        other_devices = set()

        for repo in all_remotes:
            origin_heads = [r for r in repo.refs if str(r).startswith("origin/")]

            for branch in origin_heads:
                tree = repo.tree(str(branch))
                blobs = collect_blobs(tree=tree)
                for blob in blobs:
                    if blob.path.endswith(".epc"):
                        s = "{}:{}:{}".format(
                            repo.remotes.origin.url,
                            str(branch)[len("origin/") :],
                            blob.path,
                        )
                        if s not in all_devices_strings:
                            other_devices.add(s)

        other_devices = sorted(other_devices)
        if len(other_devices) > 0:
            print()
            print("Other devices available from the referenced repositories:")
            for device in other_devices:
                print("    {}".format(device))

        # TODO: workaround for 'bug' in gitpython
        #       https://epc-phab.exana.io/T407
        #       https://github.com/gitpython-developers/GitPython/issues/546
        for repo in all_repos:
            repo.git.clear_cache()

        for dir in dirs:
            # http://bugs.python.org/issue26660
            shutil.rmtree(dir, onerror=del_rw)


class LoadJsonFiles(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        if getattr(namespace, self.dest) is None:
            setattr(namespace, self.dest, [])

        for value in values:
            getattr(namespace, self.dest).append(
                (
                    value.name,
                    json.loads(value.read(), object_pairs_hook=collections.OrderedDict),
                )
            )

        setattr(namespace, self.dest + "_path", os.path.abspath(value.name))


def parse_args(args):
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--device-file",
        "-d",
        type=argparse.FileType("r"),
        action=LoadJsonFiles,
        dest="device_files",
        nargs="+",
        required=True,
    )
    parser.add_argument("--output-directory", "-o", default=os.getcwd())
    parser.add_argument("--dry-run", "-n", action="store_true")
    parser.add_argument("--group", "-g", dest="groups", action="append", required=True)

    return parser.parse_args(args)


def main(args=None, device_files=None, output_directory=None, groups=None):
    if args is None:
        args = sys.argv[1:]

    if device_files is not None:
        for device in device_files:
            args.extend(["-d", device])

    if output_directory is not None:
        args.extend(["-o", output_directory])

    if groups is not None:
        for group in groups:
            args.extend(["-g", group])

    args = parse_args(args=args)

    for name, devices in args.device_files:
        print("Processing {}".format(name))
        collect(
            devices=devices,
            output_directory=args.output_directory,
            dry_run=args.dry_run,
            groups=args.groups,
            device_path=args.device_files_path,
        )


if __name__ == "__main__":
    os.environ["PATH"] = os.pathsep.join(
        (os.environ["PATH"], os.path.join("c:/", "Program Files", "Git", "bin"))
    )

    sys.exit(main())



from PyQt5 import QtWidgets
from PyQt5.QtCore import pyqtProperty, Qt
from PyQt5.QtGui import QFontMetrics

import epyqlib.overlaylabel_ui
import epyqlib.utils.qt


# See file COPYING in this source tree
__copyright__ = "Copyright 2018, EPC Power Corp."
__license__ = "GPLv2+"


styles = {
    "red": "background-color: rgba(255, 255, 255, 0);" "color: rgba(255, 85, 85, 25);",
    "blue": "background-color: rgba(255, 255, 255, 0);" "color: rgba(85, 85, 255, 25);",
}


class OverlayLabel(QtWidgets.QWidget):
    def __init__(self, parent=None, in_designer=False):
        super().__init__(parent=parent)

        self.in_designer = in_designer

        self._width_ratio = 0.8
        self._height_ratio = 0.8

        self.setStyleSheet(styles["red"])

        self.setAttribute(Qt.WA_TransparentForMouseEvents)

        self.ui = epyqlib.overlaylabel_ui.Ui_Form()
        self.ui.setupUi(self)

    @pyqtProperty(str)
    def text(self):
        self.ui.label.text()

    @text.setter
    def text(self, text):
        self.ui.label.setText(text)

    @pyqtProperty(float)
    def width_ratio(self):
        return self._width_ratio

    @width_ratio.setter
    def width_ratio(self, value):
        self._width_ratio = value

    @pyqtProperty(float)
    def height_ratio(self):
        return self._height_ratio

    @height_ratio.setter
    def height_ratio(self, value):
        self._height_ratio = value

    def resizeEvent(self, event):
        QtWidgets.QWidget.resizeEvent(self, event)

        self.update_overlay_size(event.size())

    def update_overlay_size(self, size):
        text = self.ui.label.text()
        if not text:
            text = "-"
        font = self.ui.label.font()
        font.setPixelSize(1000)
        metric = QFontMetrics(font)
        rect = metric.boundingRect(text)

        pixel_size_width = (
            font.pixelSize() * (size.width() * self.width_ratio) / rect.width()
        )

        pixel_size_height = (
            font.pixelSize() * (size.height() * self.height_ratio) / rect.height()
        )

        self.ui.label.setStyleSheet(
            "font-size: {}px; font-weight: bold".format(
                round(min(pixel_size_width, pixel_size_height))
            )
        )



#!/usr/bin/env python3

import os
import subprocess

try:
    hash = subprocess.check_output(
        [
            os.path.join("c:/", "Program Files", "Git", "bin", "git.exe"),
            "rev-parse",
            "HEAD",
        ]
    )
except FileNotFoundError:
    hash = subprocess.check_output(["git", "rev-parse", "HEAD"])

out_file = "revision.py"

with open(out_file, "w") as file:
    file.write("hash = '{}'\n".format(hash.decode("utf-8").rstrip()))



#!/usr/bin/env python
"""
Read TI COFF file into a python object
"""


from __future__ import print_function

from builtins import bytes, object, range
from array import array
from collections import namedtuple
from optparse import OptionParser
from struct import unpack, calcsize


# See file COPYING in this source tree
# See file ticoff.asi_license.txt in this source tree
__copyright__ = "\n".join(
    [
        'Copyright 2011 "Eliot Blennerhassett" <eblennerhassett@audioscience.com>',
        # https://gist.github.com/eliotb/1073231
        "Copyright 2016, EPC Power Corp."
        # Added Python3 support and a few tweaks
    ]
)
__license__ = "GPLv2+"


def read_struct(file, format):
    """read struct data formatted according to format"""
    datalen = calcsize(format)
    data = file.read(datalen)
    if len(data) != datalen:
        raise EOFError
    ret = unpack(format, data)
    if len(ret) < 2:
        ret = ret[0]
    return ret


def read_cstr(file):
    """read zero terminated c string from file"""
    output = ""
    while True:
        char = file.read(1)
        if len(char) == 0:
            raise RuntimeError("EOF while reading cstr")
        char = chr(bytes(char)[0])
        if char == "\0":
            break
        output += char
    return output


# From spraao8.pdf table 7
class Section(
    namedtuple(
        "Section",
        "name, virt_size, virt_addr, raw_data_size, raw_data_ptr, relocation_ptr,"
        + "linenum_ptr, linenum_count, reloc_count, flags, reserved, mem_page, data",
    )
):

    __slots__ = ()  # prevent creation of instance dictionaries
    section_fmt = "<8s9L2H"
    section_flags = [
        (0x00000001, "STYP_DSECT"),
        (0x00000002, "STYP_NOLOAD"),
        (0x00000004, "STYP_GROUPED"),
        (0x00000008, "STYP_PAD"),
        (0x00000010, "STYP_COPY"),
        (0x00000020, "STYP_TEXT"),
        (0x00000040, "STYP_DATA"),
        (0x00000080, "STYP_BSS"),
        (0x00000100, "STYP_100"),
        (0x00000200, "STYP_200"),
        (0x00000200, "STYP_400"),
        (0x00000800, "STYP_800"),
        (0x00001000, "STYP_BLOCK"),
        (0x00002000, "STYP_PASS"),
        (0x00004000, "STYP_CLINK"),
        (0x00008000, "STYP_VECTOR"),
        (0x00010000, "STYP_PADDED"),
    ]

    @property
    def is_zero_filled(self):
        return (self.data is not None) and not any([d != "\0" for d in self.data])

    def fill_value(self):
        if self.data is None:
            return None

        if len(self.data) % 4:
            return None  # only handle 32 bit fill

        a = array("I", self.data)
        if all([d == a[0] for d in a]):
            return a[0]
        else:
            return None

    @property
    def is_loadable(self):
        return not (self.flags & 0x00000010 or self.data is None)

    def __str__(self):
        s = "Section %s" % self.name
        if self.flags:
            s += " Flags ="
            for f in Section.section_flags:
                if self.flags & f[0]:
                    s = s + " " + f[1]
        for fv in self._asdict().items():
            if fv[0] == "data" or fv[0] == "name":
                continue
            s = s + "\n\t%s = 0x%X" % fv
        return s


class Symbol(
    namedtuple(
        "Symbol",
        "name, value, section_number, reserved," "storage_class, number_of_aux_entries",
    )
):

    __slots__ = ()  # prevent creation of instance dictionaries
    symbol_fmt = "<8sLhHcc"
    symbol_auxiliary_format = "<LHH10c"
    symbol_flags = [
        (0, "C_NULL"),
        (1, "C_AUTO"),
        (2, "C_EXT"),
        (3, "C_STAT"),
        (4, "C_REG"),
        (5, "C_EXTREF"),
        (6, "C_LABEL"),
        (7, "C_ULABEL"),
        (8, "C_MOS"),
        (9, "C_ARG"),
        (10, "C_STRTAG"),
        (11, "C_MOU"),
        (12, "C_UNTAG"),
        (13, "C_TPDEF"),
        (14, "C_USTATIC"),
        (15, "C_ENTAG"),
        (16, "C_MOE"),
        (17, "C_REGPARM"),
        (18, "C_FIELD"),
        (19, "C_UEXT"),
        (20, "C_STATLAB"),
        (21, "C_EXTLAB"),
        (27, "C_VARARG"),
        (100, "C_BLOCK"),
        (101, "C_FCN"),
        (102, "C_EOS"),
        (103, "C_FILE"),
        (104, "C_LINE"),
    ]


class Coff(object):
    Header = namedtuple(
        "Header",
        "machine_type, section_count, time_date, symbol_table_ptr, "
        + "symbol_count, optional_header_size, flags, target_id",
    )
    header_fmt = "<2H3L3H"

    OptionalHeader = namedtuple(
        "OptionalHeader",
        "magic, version, exe_size, init_data_size, uninit_data_size, entry_point, exe_start, init_start",
    )
    optheader_fmt = "<2H6L"

    def __init__(self, filename=None):
        self.header = None
        self.optheader = None
        self.sections = []
        if filename is not None:
            self.from_file(filename)

    def from_file(self, name):
        with open(name, "rb") as f:
            self.from_stream(f)

    def from_stream(self, f):
        self.header = self.Header(*read_struct(f, self.header_fmt))
        self.optheader = self.OptionalHeader(*read_struct(f, self.optheader_fmt))
        self.sections = []
        for i in range(self.header.section_count):
            section = Section(*read_struct(f, Section.section_fmt), data=None)
            section = section._replace(name=self.symname(f, section.name))
            if section.raw_data_ptr and section.raw_data_size:
                here = f.tell()
                f.seek(section.raw_data_ptr)
                # TODO: `2 *` is hard coded to handle the 2-bytes per
                #       address scenario.  This should obviously be
                #       detected somehow, unless it is always correct.
                data = bytes(f.read(2 * section.raw_data_size))
                f.seek(here)
                section = section._replace(data=data)
            self.sections.append(section)

        self.symbols = []
        f.seek(self.header.symbol_table_ptr)
        for i in range(self.header.symbol_count):
            symbol = Symbol(*read_struct(f, Symbol.symbol_fmt))
            try:
                symbol = symbol._replace(name=self.symname(f, symbol.name))
            except UnicodeDecodeError:
                # TODO: not sure what to do with these
                pass
            symbol = symbol._replace(
                number_of_aux_entries=symbol.number_of_aux_entries[0]
            )
            try:
                symbol = symbol._replace(
                    storage_class=next(
                        f
                        for f in symbol.symbol_flags
                        if symbol.storage_class[0] == f[0]
                    )
                )
            except StopIteration:
                print("bad {}".format(symbol))
            self.symbols.append(symbol)

            # TODO: this breaks by hitting EOF but it seems we would need
            #       to read the aux entries somewhere
            # if symbol.number_of_aux_entries:
            #     read_struct(f, Symbol.symbol_auxiliary_format)

        stack_section_number = next(
            i for i, s in enumerate(self.sections) if s.name == ".stack"
        )
        self.variables = [
            str(s)
            for s in self.symbols
            if isinstance(s.name, str)
            and s.name.startswith("_")
            and s.section_number == stack_section_number
        ]

        self.entry_point = self.optheader.entry_point
        self.sections.sort(key=lambda s: s.virt_addr)

    @property
    def loadable_sections(self):
        return [s for s in self.sections if s.is_loadable]

    def string_table_entry(self, f, offset):
        here = f.tell()
        seek = self.header.symbol_table_ptr + self.header.symbol_count * 18 + offset
        f.seek(seek)
        s = read_cstr(f)
        f.seek(here)
        return s

    def symname(self, f, value):
        parts = unpack("<2L", value)
        if parts[0] == 0:
            return self.string_table_entry(f, parts[1])
        else:
            return str(value.decode("ascii").rstrip("\0"))

    def __str__(self):
        s = "TI coff file"
        if self.header is not None:
            for fv in self.header._asdict().items():
                s = s + "\n\t%s = 0x%X" % fv
        if self.optheader is not None:
            s += "\nOptional Header"
            for fv in self.optheader._asdict().items():
                s = s + "\n\t%s = 0x%X" % fv
        return s


if __name__ == "__main__":
    parser = OptionParser()
    parser.add_option(
        "-d",
        "--dump_data",
        action="store_true",
        dest="dump",
        default=False,
        help="dump section data",
    )

    options, args = parser.parse_args()

    c = Coff(args[0])
    print(c)

    for section in c.sections:
        print(section)
        if section.data is not None and options.dump:
            print(repr(section))



from unittest.mock import MagicMock

import pytest
from twisted.internet.defer import ensureDeferred

from epyqlib.tabs.files.files_manager import FilesManager
from epyqlib.tabs.files.files_controller import FilesController
from epyqlib.tabs.files.filesview import FilesView

# noinspection PyUnresolvedReferences
from epyqlib.tests.utils.test_fixtures import temp_dir


@pytest.inlineCallbacks
@pytest.mark.skip(reason="Could break at any time if database changes")
def test_get_associations():
    view = MagicMock(spec=FilesView)
    controller = FilesController(view)

    deferred = ensureDeferred(controller.get_inverter_associations("0"))
    results = yield deferred

    view.attach_row_to_parent.assert_called()


@pytest.mark.skip(reason="Just for local testing")
@pytest.inlineCallbacks
def test_get_file(temp_dir):
    view = MagicMock(spec=FilesView)
    controller = FilesController(view)

    hash = "52e2678f71a591e9e0edfcf249ff07ae"

    controller.cache_manager = FilesManager(temp_dir)

    assert controller.cache_manager.has_hash(hash) is False

    output = yield ensureDeferred(controller.download_file(hash))
    # assert output is None

    assert controller.cache_manager.has_hash(hash) is True



#!/usr/bin/env python3

# TODO: """DocString if there is one"""

import functools
import io
import os

from PyQt5 import uic
from PyQt5.QtCore import pyqtProperty, QFile, QFileInfo, QTextStream, QTimer
from PyQt5.QtGui import QColor
from PyQt5.QtWidgets import QLabel, QHBoxLayout, QWidget

# See file COPYING in this source tree
__copyright__ = "Copyright 2016, EPC Power Corp."
__license__ = "GPLv2+"


class HmiDialog(QWidget):
    def __init__(self, parent=None, in_designer=False):
        QWidget.__init__(self, parent=parent)

        self.in_designer = in_designer

        ui = os.path.join(QFileInfo.absolutePath(QFileInfo(__file__)), "hmidialog.ui")

        # TODO: CAMPid 9549757292917394095482739548437597676742
        if not QFileInfo(ui).isAbsolute():
            ui_file = os.path.join(QFileInfo.absolutePath(QFileInfo(__file__)), ui)
        else:
            ui_file = ui
        ui_file = QFile(ui_file)
        ui_file.open(QFile.ReadOnly | QFile.Text)
        ts = QTextStream(ui_file)
        sio = io.StringIO(ts.readAll())
        self.ui = uic.loadUi(sio, self)

        self.ok_action = None
        self.cancel_action = None
        self.ui.ok_button.clicked.connect(self.ok)
        self.ui.cancel_button.clicked.connect(self.cancel)

        self.label.setWordWrap(True)

    def focus(self, label, ok_action=None, cancel_action=None, enable_delay=1500):
        self.ok_action = ok_action
        self.ui.ok_button.setVisible(ok_action is not None)
        self.cancel_action = cancel_action
        self.ui.cancel_button.setVisible(cancel_action is not None)

        self.ui.label.setText(label)

        if enable_delay > 0:
            self.enable_buttons(False)
        else:
            self.enable_buttons(True)

        self.parent().setCurrentWidget(self)

        if enable_delay > 0:
            QTimer.singleShot(
                enable_delay, functools.partial(self.enable_buttons, enable=True)
            )

    def enable_buttons(self, enable):
        self.ui.ok_button.setEnabled(enable)
        self.ui.cancel_button.setEnabled(enable)

    def ok(self):
        self.ok_action()

    def cancel(self):
        self.cancel_action()


if __name__ == "__main__":
    import sys

    print("No script functionality here")
    sys.exit(1)  # non-zero is a failure



import pathlib
import typing

import PyQt5.uic
import attr
from PyQt5.QtGui import QFont
from PyQt5.QtWidgets import QLineEdit, QPushButton, QLabel, QDialogButtonBox
from epyqlib.tabs.files.cognito import CognitoException

# noinspection PyUnreachableCode
if typing.TYPE_CHECKING:
    from epyqlib.tabs.files.aws_login_manager import AwsLoginManager

Ui, UiBase = PyQt5.uic.loadUiType(
    pathlib.Path(__file__).with_suffix(".ui"),
)


@attr.s
class LoginDialog(UiBase):
    login_manager: "AwsLoginManager" = attr.ib()
    ui = attr.ib(factory=Ui)

    invalid_input = False

    def __attrs_post_init__(self):
        super().__init__()
        self.setup_ui()

    @classmethod
    def build(cls):
        instance = cls()
        instance.setup_ui()

        return instance

    # noinspection PyAttributeOutsideInit
    def setup_ui(self):
        self.ui.setupUi(self)

        self._bind()
        self._fiddle_title_label()
        self._clear_error_message()

    def _fiddle_title_label(self):
        self.ui.spacer_label.setText(self.ui.big_label.text())

        reference_font = QFont()
        reference_font.setPointSize(reference_font.pointSize() * 2)
        reference_font.setWeight(QFont.Black)

        self.ui.big_label.setFont(reference_font)

        reference_font.setPointSize(reference_font.pointSize() * 2 / 3)
        self.ui.spacer_label.setFont(reference_font)

    def _bind(self):
        # self.dialog: QDialog = self.ui.root_dialog

        self.username: QLineEdit = self.ui.username
        self.password: QLineEdit = self.ui.password

        self.error_message: QLabel = self.ui.lbl_error_message

        self.btn_cancel: QPushButton = self.ui.button_box.button(
            QDialogButtonBox.StandardButton.Cancel
        )
        self.btn_login: QPushButton = self.ui.button_box.button(
            QDialogButtonBox.StandardButton.Ok
        )
        self.btn_login.setText("Login")

        self.username.textChanged.connect(self._text_changed)
        self.password.textChanged.connect(self._text_changed)

        self.btn_cancel.clicked.connect(self._cancel_clicked)
        self.btn_login.clicked.connect(self._login_clicked)

    def _cancel_clicked(self):
        self.reject()

    def _login_clicked(self):
        self._clear_error_message()

        username = self.username.text()
        password = self.password.text()
        try:
            self.login_manager.authenticate(username, password)
        except CognitoException as e:
            self._show_error_message(e.message)
            self.invalid_input = True
            self.btn_login.setDisabled(True)
            return

        self.accept()

    def _clear_error_message(self):
        self.error_message.setText(None)

    def _show_error_message(self, err: str):
        self.error_message.setText(f'<font color="red">{err} Please try again.</font>')

    def _text_changed(self, new_text: str):
        if self.invalid_input:
            self.invalid_input = False
            self.btn_login.setEnabled(True)



import PyQt5.QtCore

# See file COPYING in this source tree
__copyright__ = "Copyright 2016, EPC Power Corp."
__license__ = "GPLv2+"


class NotFoundError(Exception):
    pass


class MultipleFoundError(Exception):
    pass


class Signals(PyQt5.QtCore.QObject):
    child_added = PyQt5.QtCore.pyqtSignal("PyQt_PyObject", int)
    child_added_complete = PyQt5.QtCore.pyqtSignal("PyQt_PyObject")
    child_removed = PyQt5.QtCore.pyqtSignal(
        "PyQt_PyObject",
        "PyQt_PyObject",
        int,
    )
    child_removed_complete = PyQt5.QtCore.pyqtSignal("PyQt_PyObject")


class TreeNode:
    def __init__(self, tx=False, parent=None, children=None):
        self.last = None

        self.tx = tx

        self.tree_parent = None
        self.set_parent(parent)

        self.pyqt_signals = Signals()

        # TODO: this isn't a good way to handle predefined children
        #       in an inherited class
        if children is None:
            if hasattr(self, "children"):
                children = self.children

        if children is None:
            self.children = []
        else:
            children = self.children
            self.children = []
            for child in children:
                self.append_child(child)

    def set_parent(self, parent):
        self.tree_parent = parent
        if self.tree_parent is not None:
            self.tree_parent.append_child(self)

    def insert_child(self, i, child):
        self.children.insert(i, child)
        child.tree_parent = self
        self.pyqt_signals.child_added.emit(child, i)
        self.pyqt_signals.child_added_complete.emit(child)

    def append_child(self, child):
        self.children.append(child)
        child.tree_parent = self
        self.pyqt_signals.child_added.emit(child, len(self.children) - 1)
        self.pyqt_signals.child_added_complete.emit(child)

    def child_at_row(self, row):
        if row < len(self.children):
            return self.children[row]
        else:
            return None

    def row_of_child(self, child):
        for i, item in enumerate(self.children):
            if item is child:
                return i
        return -1

    def remove_child(self, row=None, child=None):
        if child is None:
            child = self.children[row]
        elif row is None:
            row = self.children.index(child)

        tree_parent = child.tree_parent

        child.parent = None
        child.tree_parent = None
        self.children.remove(child)

        self.pyqt_signals.child_removed.emit(tree_parent, child, row)
        self.pyqt_signals.child_removed_complete.emit(child)

        return True

    def traverse(self, call_this, payload=None, internal_nodes=False):
        if internal_nodes or len(self.children) == 0:
            call_this(self, payload)

        for child in self.children:
            child.traverse(call_this, payload, internal_nodes=internal_nodes)

    def leaves(self):
        leaves = []
        self.traverse(
            call_this=lambda node, payload: payload.append(node),
            payload=leaves,
            internal_nodes=False,
        )

        return leaves

    def find_root(self):
        root = self

        while root.tree_parent is not None:
            root = root.tree_parent

        return root

    def ancestors(self):
        node = self

        while node.tree_parent is not None:
            node = node.tree_parent
            yield node

    def children_by_attribute(self, value, name):
        return [
            child
            for child in self.children
            if hasattr(child, name) and getattr(child, name) == value
        ]

    def __len__(self):
        return len(self.children)

    def nodes_by_attribute(self, attribute_value, attribute_name, raise_=True):
        def matches(node):
            if not hasattr(node, attribute_name):
                return False

            return getattr(node, attribute_name) == attribute_value

        nodes = self.nodes_by_filter(filter=matches)

        if len(nodes) == 0 and raise_:
            raise NotFoundError(
                """Attribute '{}' with value '{}' not found""".format(
                    attribute_name,
                    attribute_value,
                )
            )

        return nodes

    def nodes_by_filter(self, filter, collection=None):
        def visit(node, matches):
            if filter(node):
                if isinstance(matches, set):
                    matches.add(node)
                if isinstance(matches, list):
                    matches.append(node)

        if collection is None:
            collection = set()

        self.traverse(call_this=visit, payload=collection, internal_nodes=True)

        return collection

    def recursively_remove_children(self):
        nodes = []
        self.traverse(
            call_this=lambda node, payload: payload.append(node),
            payload=nodes,
            internal_nodes=True,
        )

        nodes = [node for node in nodes if node != self]

        bottom_up = sorted(
            nodes,
            key=lambda node: (node.tree_parent.uuid, node.uuid),
        )

        for node in bottom_up:
            node.tree_parent.remove_child(child=node)

        return nodes

    def child_by_name(self, name):
        sentinel = object()

        children = [
            child for child in self.children if getattr(child, "name", sentinel) == name
        ]

        if len(children) == 0:
            raise NotFoundError(f"Child with name {name!r} not found")

        if len(children) > 1:
            raise MultipleFoundError(
                f"Multiple children found with name {name!r}",
            )

        (child,) = children

        return child

    def descendent(self, *names):
        node = self
        for name in names:
            node = node.child_by_name(name)

        return node


if __name__ == "__main__":
    import sys

    print("No script functionality here")
    sys.exit(1)  # non-zero is a failure



import collections

import attr


def create_helpers(cls):
    lengths = cls.lengths()
    offsets = (sum(lengths[:i]) for i in range(len(lengths)))
    cls_offsets = cls(*offsets)
    cls_slices = cls(
        *(slice(o, o + l) for l, o in zip(lengths, attr.astuple(cls_offsets)))
    )

    sf = "{{:0{}b}}".format(sum(lengths))

    def unpack(n):
        sn = sf.format(n)
        return cls(*(int(sn[s], 2) for s in attr.astuple(cls_slices)))

    def pgn(id):
        strings = (
            "{:0{}b}".format(n, l)
            for n, l, b in zip(attr.astuple(id), lengths, cls.pgns())
            if b
        )
        return int("".join(strings), 2)

    def pack(id):
        strings = ("{:0{}b}".format(n, l) for n, l in zip(attr.astuple(id), lengths))
        return int("".join(strings), 2)

    return unpack, pgn, pack


def add_helpers(cls):
    cls.unpack, cls.pgn, cls.pack = create_helpers(cls)

    return cls


class IncorrectFormat(Exception):
    pass


@add_helpers
@attr.s
class Id:
    priority = attr.ib(metadata={"length": 3, "pgn": False})
    extended_data_page = attr.ib(metadata={"length": 1, "pgn": True})
    data_page = attr.ib(metadata={"length": 1, "pgn": True})
    pdu_format = attr.ib(metadata={"length": 8, "pgn": True})
    pdu_specific = attr.ib(metadata={"length": 8, "pgn": True})
    source_address = attr.ib(metadata={"length": 8, "pgn": False})

    @classmethod
    def lengths(cls):
        return tuple(f.metadata["length"] for f in attr.fields(cls))

    @classmethod
    def offsets(cls):
        return tuple(sum(cls.lengths()[:i]) for i in range(len(attr.fields(cls))))

    @classmethod
    def slices(cls):
        return tuple(slice(o, o + l) for l, o in zip(cls.lengths(), cls.offsets()))

    @classmethod
    def pgns(cls):
        return tuple(f.metadata["pgn"] for f in attr.fields(cls))

    def is_proprietary_a(self):
        return all(
            (
                self.extended_data_page == 0,
                self.data_page == 0,
                self.pdu_format == 239,
            )
        )

    def is_proprietary_a2(self):
        return all(
            (
                self.extended_data_page == 0,
                self.data_page == 1,
                self.pdu_format == 239,
            )
        )

    def is_proprietary_b(self):
        return all(
            (
                self.extended_data_page == 0,
                self.data_page == 0,
                self.pdu_format == 255,
            )
        )

    def is_iso15765_3(self):
        return all(
            (
                self.extended_data_page == 1,
                self.data_page == 1,
            )
        )

    def is_pdu1(self):
        return 0 <= self.pdu_format <= 239

    def is_pdu2(self):
        return 240 <= self.pdu_format <= 255

    @property
    def destination_address(self):
        if not self.is_pdu1():
            raise IncorrectFormat(
                "Requested destination address for non-PDU1 identifer"
            )

        return self.pdu_specific

    @destination_address.setter
    def destination_address(self, value):
        if not self.is_pdu1():
            raise IncorrectFormat(
                "Attempted to set destination address for non-PDU1 identifer"
            )

        self.pdu_specific = value

    @property
    def group_extension(self):
        if not self.is_pdu2():
            raise IncorrectFormat("Requested group extension for non-PDU2 identifer")

        return self.pdu_specific

    @group_extension.setter
    def group_extension(self, value):
        if not self.is_pdu2():
            raise IncorrectFormat(
                "Attempted to set group extension for non-PDU2 identifer"
            )

        self.pdu_specific = value


def example():
    print(Id(*Id.lengths()))
    print(Id(*Id.offsets()))
    print(Id(*Id.slices()))

    iid = 0x0CFFC2F6
    sid = "{:029b}".format(iid)
    print(sid)
    print(" ".join(sid[s] for s in Id.slices()))
    id = Id(*(int(sid[s], 2) for s in Id.slices()))
    print(id)

    print(Id.unpack(iid))

    print()
    # grep 'ID=' .sym | sed 's/ID=\([0-9A-Fa-f]\+\)h.*/    0x\1,/'
    ids = [
        0x1DEFF741,
        0x0CEFF741,
        0x1DEF41F7,
        0x18FFD0F7,
        0x18FFD1F7,
        0x18FFD2F7,
        0x1CFFC0F7,
        0x0CFFCAF7,
        0x18FFC4F7,
        0x0CFFC3F7,
        0x18FFCBF7,
        0x0CFFC2F7,
        0x1CFFC7F7,
        0x1CFFCCF7,
        0x0CFFC1F7,
        0x18FFD3F7,
        0x18FFD4F7,
        0x1FFFFFF7,
        0x1FFFFFFF,
    ]

    for id in ids:
        cls_id = Id.unpack(id)
        print(
            "PGN: 0x{pgn: 7X} {pgn: 7d} 0x{id:08x}: {cls}".format(
                pgn=Id.pgn(cls_id), id=id, cls=cls_id
            )
        )

    print()
    i_id = 247
    st_id = 65

    process_to_i = Id(
        priority=3,
        extended_data_page=0,
        data_page=0,
        pdu_format=239,
        pdu_specific=i_id,
        source_address=st_id,
    )
    setup_to_i = Id(
        priority=7,
        extended_data_page=0,
        data_page=1,
        pdu_format=239,
        pdu_specific=i_id,
        source_address=st_id,
    )
    process_from_i = Id(
        priority=3,
        extended_data_page=0,
        data_page=0,
        pdu_format=239,
        pdu_specific=st_id,
        source_address=i_id,
    )
    setup_from_i = Id(
        priority=7,
        extended_data_page=0,
        data_page=1,
        pdu_format=239,
        pdu_specific=st_id,
        source_address=i_id,
    )

    ids = collections.OrderedDict(
        (
            ("process_to_i", process_to_i),
            ("setup_to_i", setup_to_i),
            ("process_from_i", process_from_i),
            ("setup_from_i", setup_from_i),
        )
    )

    print()
    for name, id in ids.items():
        raw_id = Id.pack(id)
        print("{:>20s}: PGN: {: 7d} 0x{:08x} {}".format(name, Id.pgn(id), raw_id, id))



import pathlib

import click
import PyQt5.uic


@click.command()
@click.option(
    "--ui",
    "ui_paths",
    multiple=True,
    type=click.Path(exists=True, dir_okay=False),
)
@click.option(
    "--directory",
    "--dir",
    "directories",
    default=["."],
    multiple=True,
    type=click.Path(exists=True, file_okay=False),
)
@click.option("--suffix", default="_ui")
@click.option("--encoding", default="utf-8")
def cli(ui_paths, directories, suffix, encoding):
    ui_paths = [pathlib.Path(path) for path in ui_paths]

    for directory in directories:
        path = pathlib.Path(directory)
        found_paths = path.rglob("*.ui")
        ui_paths.extend(found_paths)

    for path in ui_paths:
        in_path = path
        out_path = path.with_name(f"{path.stem}{suffix}.py")

        click.echo(f"Converting: {in_path} -> {out_path}")
        with open(out_path, "w", encoding=encoding) as out_file:
            PyQt5.uic.compileUi(in_path, out_file)



import operator

import epyqlib._build
import epyqlib._version


__version__, __sha__ = operator.itemgetter("version", "full-revisionid")(
    epyqlib._version.get_versions(),
)

__version_tag__ = "v{}-{}".format(__version__, __sha__)
__build_tag__ = epyqlib._build.job_id



from PyQt5 import QtDesigner
import epyqlib.searchbox
import epyqlib.abstractpluginclass

# See file COPYING in this source tree
__copyright__ = "Copyright 2016, EPC Power Corp."
__license__ = "GPLv2+"


class SearchBoxPlugin(epyqlib.abstractpluginclass.AbstractPlugin):
    def __init__(self, parent=None):
        super().__init__(parent=parent)

        self._group = "EPC - General"
        self._init = epyqlib.searchbox.SearchBox
        self._module_path = "epyqlib.searchbox"
        self._name = "SearchBox"



#!/usr/bin/env python3

# TODO: """DocString if there is one"""

import functools

from PyQt5 import QtWidgets

import epyqlib.listmenu
import epyqlib.listselect_ui
import epyqlib.utils.qt


# See file COPYING in this source tree
__copyright__ = "Copyright 2016, EPC Power Corp."
__license__ = "GPLv2+"


class ListSelect(QtWidgets.QWidget):
    def __init__(self, parent=None, in_designer=False, action=None):
        super().__init__(parent=parent)

        self.in_designer = in_designer

        self.ui = epyqlib.listselect_ui.Ui_Form()
        self.ui.setupUi(self)

        self.action = action
        self.items = {}
        self.ui.accept_button.clicked.connect(self.accept)
        self.ui.cancel_button.clicked.connect(functools.partial(self.exit, value=None))

    def focus(self, value, action, items, label=""):
        self.items = items
        root = epyqlib.listmenu.Node(text=label)
        model = epyqlib.listmenu.ListMenuModel(root=root)
        self.ui.menu_view.setModel(model)

        selected = None

        for key, item_value in sorted(self.items.items()):
            node = epyqlib.listmenu.Node(text=item_value, action=lambda: None)
            root.append_child(node)

            if key == int(value):
                selected = node

        if selected is not None:
            self.ui.menu_view.select_node(selected)

        self.action = action
        parent = self.parent()
        if hasattr(parent, "setCurrentWidget"):
            parent.setCurrentWidget(self)
        focused_widget = self.focusWidget()
        if focused_widget is not None:
            focused_widget.clearFocus()

    def accept(self):
        selected = self.ui.menu_view.selected_text()
        found_key = None
        for key, value in self.items.items():
            if value == selected:
                found_key = key
                break

        self.exit(value=found_key)

    def exit(self, value):
        self.action(value=value)


if __name__ == "__main__":
    import sys

    print("No script functionality here")
    sys.exit(1)  # non-zero is a failure



import os
import pkg_resources
import sys
import textwrap

import click


@click.command()
@click.option("--dist", "--distribution", "-d", "distribution", required=True)
@click.option("--out", "-o", type=click.File("w"), default=sys.stdout)
def cli(distribution, out):
    import pkg_resources

    version = pkg_resources.get_distribution(distribution).version

    out.write(
        textwrap.dedent(
            """\
        # This file is automatically generated by {me}
        __version__ = '{version}'
        __sha__ = '{sha}'
        __revision__ = '{revision}'
    """.format(
                me=os.path.splitext(os.path.basename(sys.argv[0]))[0],
                version=version,
                sha="",
                revision="",
            )
        )
    )



import hashlib
import inspect
import json
import os
import shutil
from datetime import datetime
from enum import Enum
from os import path
from typing import Union, List, Callable, Coroutine

import attr

from epyqlib.tabs.files.files_utils import ensure_dir

NewLogListener = Callable[[str], Coroutine]


@attr.s(slots=True, auto_attribs=True)
class PendingLog:
    hash: str
    filename: str
    build_id: str = None
    serial_number: str = None
    username = "Current User"
    notes = ""
    timestamp = datetime.now()


class LogManager:
    _instance: "LogManager" = None
    _tag = "[Log Manager]"

    class EventType(Enum):
        log_generated = 1
        log_synced = 2

    def __init__(self, files_dir: str):
        # if LogManager._instance is not None:
        #     raise Exception("LogManager being created instead of using singleton")

        self._cache_dir = path.join(files_dir, "pending_logs")
        ensure_dir(self._cache_dir)

        self._pending_logs_file = path.join(files_dir, "pending-logs.json")
        self._pending_logs: List[PendingLog] = []

        self._listeners: List[NewLogListener] = []

    @staticmethod
    def get_instance():
        if LogManager._instance is None:
            raise Exception("LogManager being used before initialized")
        return LogManager._instance

    @staticmethod
    def init(files_dir: str):
        LogManager._instance = LogManager(files_dir)
        LogManager._instance._read_pending_log_file()
        return LogManager._instance

    async def add_pending_log(self, file_path: str, build_id: str, serial_number: str):
        basename = os.path.basename(file_path)
        hash = self._md5(file_path)
        shutil.copy2(file_path, path.join(self._cache_dir, hash))

        new_log = PendingLog(hash, basename, build_id, serial_number)
        self._pending_logs.append(new_log)
        self._save_pending_log_file()

        await self._notify_listeners(new_log)

    def _save_pending_log_file(self):
        # Don't save anything if there's nothing to save
        if len(self._pending_logs) == 0:
            if os.path.exists(self._pending_logs_file):
                os.unlink(self._pending_logs_file)
            return

        with open(self._pending_logs_file, "w") as file:
            data = [attr.asdict(log) for log in self._pending_logs]
            json.dump(data, file, indent=2)

    def _read_pending_log_file(self):
        if not path.exists(self._pending_logs_file):
            return

        with open(self._pending_logs_file, "r") as file:
            data: List[dict] = json.load(file)
            for log in data:
                self._pending_logs.append(PendingLog(**log))

    def get_path_to_log(self, hash: str):
        return path.join(self._cache_dir, hash)

    def get_pending_logs(self) -> List[PendingLog]:
        return self._pending_logs

    def get_next_pending_log(self) -> Union[PendingLog, None]:
        if len(self._pending_logs) > 0:
            return self._pending_logs[0]
        else:
            return None

    def remove_pending(self, log: PendingLog):
        self._pending_logs.remove(log)
        self._save_pending_log_file()

    def _md5(self, filename: str) -> str:
        md5 = hashlib.md5()
        with open(filename, "rb") as file:
            for chunk in iter(lambda: file.read(4096), b""):
                md5.update(chunk)
        return md5.hexdigest()

    def get_file_ref(self, filename: str, mode: str):
        return open(path.join(self._cache_dir, filename), mode)

    def stat(self, filename) -> os.stat_result:
        return os.stat(path.join(self._cache_dir, filename))

    ## Listener Management
    ## Listener fires when new log is added
    def add_listener(self, listener: NewLogListener):
        self._listeners.append(listener)

    def remove_listener(self, listener: NewLogListener):
        self._listeners.remove(listener)

    async def _notify_listeners(self, log: PendingLog):
        for listener in self._listeners:
            result = listener(log)
            if inspect.iscoroutine(result):
                await result



import functools
import decimal
import logging
import time

import attr
import twisted.internet.defer

import epyqlib.utils.general
import epyqlib.utils.qt

__copyright__ = "Copyright 2016, EPC Power Corp."
__license__ = "GPLv2+"


logger = logging.getLogger(__name__)


class RequestTimeoutError(TimeoutError):
    pass


class WaitForTimedOut(Exception):
    pass


async def wait_for(check, period=0.1, timeout=10, message=None):
    if message is None:
        message = f"Condition not satisfied within {timeout:.1f} seconds"

    start = time.monotonic()

    while True:
        done = await check()
        if done:
            return

        if time.monotonic() - start > timeout:
            raise WaitForTimedOut(message)

        await epyqlib.utils.twisted.sleep(period)


def ignore_cancelled(f):
    async def wrapper(*args, **kwargs):
        try:
            return await f(*args, **kwargs)
        except twisted.internet.defer.CancelledError:
            pass

    return wrapper


def ensure_deferred(f):
    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        return twisted.internet.defer.ensureDeferred(f(*args, **kwargs))

    return wrapper


def errback_dialog(f):
    @functools.wraps(f)
    async def wrapper(*args, **kwargs):
        try:
            return await f(*args, **kwargs)
        except Exception as e:
            epyqlib.utils.twisted.errbackhook(
                twisted.python.failure.Failure(e),
            )

    return wrapper


def errbackhook(error):
    epyqlib.utils.qt.custom_exception_message_box(
        brief="{}\n{}".format(error.type.__name__, error.value),
        extended=error.getTraceback(),
    )

    return error


def discard_cancelled(failure):
    if issubclass(failure.type, twisted.internet.defer.CancelledError):
        return None

    return failure


def catch_expected(error):
    if issubclass(error.type, epyqlib.utils.general.ExpectedException):
        epyqlib.utils.qt.raw_exception_message_box(
            brief=error.value.expected_message(),
            extended=error.getTraceback(),
        )

        return None

    return error


def detour_result(result, f, *args, **kwargs):
    f(*args, **kwargs)

    return result


def logit(it):
    logger.debug("logit(): ({}) {}".format(type(it), it))

    if isinstance(it, twisted.python.failure.Failure):
        it.printDetailedTraceback()


@twisted.internet.defer.inlineCallbacks
def retry(function, times, acceptable=None):
    if acceptable is None:
        acceptable = []

    remaining = times
    while remaining > 0:
        try:
            result = yield function()
        except Exception as e:
            if type(e) not in acceptable:
                raise
        else:
            twisted.internet.defer.returnValue(result)

        remaining -= 1

    raise Exception("out of retries")


def timeout_retry(function, times=3, acceptable=None):
    if acceptable is None:
        acceptable = [RequestTimeoutError]

    # d = twisted.internet.defer.Deferred()
    # d.addCallback(retry(function=function, times=times, acceptable=acceptable))

    return retry(function=function, times=times, acceptable=acceptable)


def sleep(seconds=None):
    d = twisted.internet.defer.Deferred()

    if seconds is not None:
        if isinstance(seconds, decimal.Decimal):
            seconds = float(seconds)

        twisted.internet.reactor.callLater(seconds, d.callback, None)

    return d


class InvalidAction(Exception):
    pass


@attr.s
class DeferLaterChain:
    deferred = attr.ib(
        default=attr.Factory(twisted.internet.defer.Deferred),
        init=False,
    )
    paused = attr.ib(default=False, init=False)
    canceled = attr.ib(default=False, init=False)
    active = attr.ib(default=None, init=False)

    def add_delayed_callback(self, delay, c, *args, **kwargs):
        def defer_later():
            return twisted.internet.task.deferLater(
                clock=twisted.internet.reactor,
                delay=delay,
                callable=c,
                *args,
                **kwargs,
            )

        def callback(_):
            self.active = defer_later()
            return self.active

        self.deferred.addCallback(callback)

    def add_callback(self, *args, **kwargs):
        return self.deferred.addCallback(*args, **kwargs)

    def add_both(self, *args, **kwargs):
        return self.deferred.addBoth(*args, **kwargs)

    def add_errback(self, *args, **kwargs):
        return self.deferred.addErrback(*args, **kwargs)

    addCallback = add_callback
    addBoth = add_both
    addErrback = add_errback

    def run(self, arg=None):
        self.deferred.callback(arg)

    def cancel(self):
        self.deferred.cancel()
        self.canceled = True

    def pause(self):
        if self.canceled:
            raise InvalidAction("attempted to pause while canceled")

        if self.paused:
            raise InvalidAction("attempted to pause while paused")

        self.deferred.pause()
        self.active.pause()
        self.paused = True

    def unpause(self):
        if self.canceled:
            raise InvalidAction("attempted to unpause while canceled")

        if not self.paused:
            raise InvalidAction("attempted unpause while not paused")

        self.active.unpause()
        self.deferred.unpause()
        self.paused = False


class DeferredRepeater:
    def __init__(self, f, *args, **kwargs):
        self.callable = f
        self.args = args
        self.kwargs = kwargs

        self.public_deferred = twisted.internet.defer.Deferred(canceller=self.cancel)
        self.private_deferred = None

    def repeat(self):
        self.private_deferred = self.callable(*self.args, **self.kwargs)
        self.private_deferred.addCallback(lambda _: self.repeat())
        self.private_deferred.addErrback(self.public_deferred.errback)

    def cancel(self, _=None):
        if self.private_deferred is not None:
            self.private_deferred.cancel()

    def pause(self):
        return self.private_deferred.pause()

    def unpause(self):
        return self.private_deferred.unpause()


@attr.s
class Mobius:
    f = attr.ib()
    args = attr.ib(default=())
    kwargs = attr.ib(default=attr.Factory(dict))
    repeater = attr.ib()
    public_deferred = attr.ib()

    @repeater.default
    def _(self):
        return DeferredRepeater(self.f, *self.args, **self.kwargs)

    @public_deferred.default
    def _(self):
        return twisted.internet.defer.Deferred(
            canceller=self.repeater.cancel,
        )

    def __attrs_post_init__(self):
        self.repeater.public_deferred.addErrback(self.public_deferred.errback)

    @classmethod
    def build(cls, f, *args, **kwargs):
        return cls(f=f, args=args, kwargs=kwargs)

    def add_errback(self, *args, **kwargs):
        self.public_deferred.addErrback(*args, **kwargs)

    addErrback = add_errback

    def run(self):
        self.repeater.repeat()

        return self.public_deferred

    def cancel(self):
        return self.public_deferred.cancel()

    def pause(self):
        self.public_deferred.pause()
        self.repeater.pause()

    def unpause(self):
        self.public_deferred.unpause()
        self.repeater.unpause()


@attr.s
class Action:
    f = attr.ib()
    args = attr.ib(default=())
    kwargs = attr.ib(default=attr.Factory(dict))

    def __call__(self):
        return self.f(*self.args, **self.kwargs)


@attr.s
class Event:
    action = attr.ib()
    time = attr.ib()


@attr.s
class Sequence:
    events = attr.ib(default=attr.Factory(list))
    tolerance = attr.ib(default=0.005)
    clock = attr.ib(default=time.monotonic)

    virtual_time = attr.ib(default=0, init=False)
    clock_time = attr.ib(default=0, init=False)
    deferred = attr.ib(default=None, init=False)
    paused = attr.ib(default=False, init=False)
    run_deferred = attr.ib(default=None, init=False)

    def add_delayed(self, delay, f, *args, **kwargs):
        last_time = 0
        if len(self.events) > 0:
            last_time = self.events[-1].time

        self.events.append(
            Event(
                action=Action(f=f, args=args, kwargs=kwargs),
                time=last_time + delay,
            )
        )

    def run(self, loop=False):
        self.run_deferred = self._run(loop=loop)

        return self.run_deferred

    @twisted.internet.defer.inlineCallbacks
    def _run(self, loop=False):
        self.update_time(virtual_time=0)

        run = True
        while run:
            run = loop

            base_time = self.virtual_time

            for event in self.events:
                while True:
                    delay = (base_time + event.time) - self.virtual_time
                    logger.debug(
                        "self.virtual_time: {}".format(self.virtual_time),
                    )
                    logger.debug("delay: {}".format(delay))
                    if delay > self.tolerance:
                        self.deferred = epyqlib.utils.twisted.sleep(
                            delay - (self.tolerance / 2),
                        )
                        try:
                            yield self.deferred
                        except twisted.internet.defer.CancelledError:
                            if not self.paused:
                                raise
                        self.deferred = None

                    self.update_time()

                    if not self.paused:
                        break

                    if self.paused:
                        self.deferred = sleep()
                        try:
                            yield self.deferred
                        except twisted.internet.defer.CancelledError:
                            if self.paused:
                                raise
                        finally:
                            self.deferred = None

                        self.update_time(virtual_time=self.virtual_time)

                event.action()
                self.update_time()

    def update_time(self, virtual_time=None):
        now = self.clock()

        if virtual_time is None:
            self.virtual_time += now - self.clock_time

        self.clock_time = now

    def cancel(self):
        self.run_deferred.cancel()

    def pause(self):
        self.paused = True
        if self.deferred is not None:
            self.deferred.cancel()

    def unpause(self):
        self.paused = False
        if self.deferred is not None:
            self.deferred.cancel()



import gc
import time

import attr
import pytest
import pytest_twisted
import twisted.internet.defer
import twisted.logger

import epyqlib.utils.twisted


@attr.s
class Elapsed:
    clock = attr.ib(default=time.monotonic)
    start = attr.ib()

    @start.default
    def _(self):
        return self.clock()

    def number_string(self):
        now = time.monotonic() - self.start
        return now, "T+ {:.1f}s".format(now)


@attr.s
class Result:
    time = attr.ib()
    data = attr.ib()


@attr.s
class ActionLogger:
    elapsed = attr.ib(default=attr.Factory(Elapsed))
    log = attr.ib(default=attr.Factory(list))

    expected = attr.ib(default=attr.Factory(list))

    def action(self, data):
        now, s = self.elapsed.number_string()
        print(s, data)
        self.log.append(Result(time=now, data=data))

    def check(self):
        assert len(self.expected) == len(self.log)
        for result, expected in zip(self.log, self.expected):
            assert abs(result.time - expected.time) < 0.050
            assert expected.data == result.data


@pytest.fixture(
    params=[
        pytest.param(
            "flaky",
            marks=pytest.mark.flaky(reruns=5),
        ),
    ],
)
def action_logger():
    logger = ActionLogger()
    yield logger
    logger.check()


# https://github.com/pytest-dev/pytest-twisted/issues/4#issuecomment-360888462
class Observer:
    def __init__(self):
        self.failures = []

    def __call__(self, event_dict):
        is_error = event_dict.get("isError")
        s = "Unhandled error in Deferred".casefold()
        is_unhandled = s in event_dict.get("log_format", "").casefold()

        if is_error and is_unhandled:
            self.failures.append(event_dict)

    def check(self):
        assert [] == self.failures


@pytest.fixture
def assert_no_unhandled_errbacks():
    observer = Observer()
    twisted.logger.globalLogPublisher.addObserver(observer)

    yield

    # deferred's err for being unhandled in __del__ so...  try and hope
    gc.collect()
    twisted.logger.globalLogPublisher.removeObserver(observer)

    observer.check()


@pytest_twisted.inlineCallbacks
def test_sequence_normal(action_logger, assert_no_unhandled_errbacks):
    sequence = epyqlib.utils.twisted.Sequence()
    t = 0
    delay = 0.3
    for n in range(3):
        sequence.add_delayed(delay, action_logger.action, n)
        t += delay
        action_logger.expected.append(Result(time=t, data=n))

    yield sequence.run()


@pytest_twisted.inlineCallbacks
def test_sequence_cancelled(action_logger, assert_no_unhandled_errbacks):
    sequence = epyqlib.utils.twisted.Sequence()
    t = 0
    delay = 0.3
    for n in range(3):
        sequence.add_delayed(delay, action_logger.action, n)
        t += delay
        action_logger.expected.append(Result(time=t, data=n))

    cancel_time = action_logger.expected.pop().time - (delay / 2)

    deferred = sequence.run()

    yield epyqlib.utils.twisted.sleep(cancel_time)
    sequence.cancel()

    with pytest.raises(twisted.internet.defer.CancelledError):
        yield deferred


@pytest_twisted.inlineCallbacks
def test_sequence_loop(action_logger, assert_no_unhandled_errbacks):
    sequence = epyqlib.utils.twisted.Sequence()
    t = 0
    delay = 0.3
    for repeat in (False, True, True):
        for n in range(3):
            if not repeat:
                sequence.add_delayed(delay, action_logger.action, n)
            t += delay
            action_logger.expected.append(Result(time=t, data=n))

    cancel_time = action_logger.expected[-1].time + (delay / 2)

    deferred = sequence.run(loop=True)

    yield epyqlib.utils.twisted.sleep(cancel_time)
    sequence.cancel()

    with pytest.raises(twisted.internet.defer.CancelledError):
        yield deferred


@pytest_twisted.inlineCallbacks
def test_sequence_pause(action_logger, assert_no_unhandled_errbacks):
    sequence = epyqlib.utils.twisted.Sequence()
    t = 0
    delay = 0.3
    for n in range(3):
        sequence.add_delayed(delay, action_logger.action, n)
        t += delay
        action_logger.expected.append(Result(time=t, data=n))

    pause_before = 1

    pause_time = action_logger.expected[pause_before].time - (delay / 2)
    unpause_time = pause_time + 1

    for expected in action_logger.expected[pause_before:]:
        expected.time += unpause_time - pause_time

    deferred = sequence.run()

    twisted.internet.reactor.callLater(pause_time, sequence.pause)
    twisted.internet.reactor.callLater(unpause_time, sequence.unpause)

    yield deferred


@pytest_twisted.inlineCallbacks
def test_sequence_pause_event(action_logger, assert_no_unhandled_errbacks):
    sequence = epyqlib.utils.twisted.Sequence()
    t = 0
    delay = 0.3
    for n in range(3):
        sequence.add_delayed(
            delay,
            action_logger.action,
            n,
        )
        t += delay
        action_logger.expected.append(Result(time=t, data=n))

    pause_before = 2

    pause_data = object()

    def pause():
        sequence.pause()
        action_logger.action(pause_data)

    pause_time = sequence.events[pause_before].time
    pause_event = epyqlib.utils.twisted.Event(
        time=pause_time,
        action=epyqlib.utils.twisted.Action(f=pause),
    )

    sequence.events.insert(pause_before, pause_event)
    action_logger.expected.insert(
        pause_before,
        Result(time=pause_time, data=pause_data),
    )

    pause_duration = 1
    unpause_time = pause_time + pause_duration

    for expected in action_logger.expected[pause_before + 1 :]:
        expected.time += pause_duration

    deferred = sequence.run()

    yield epyqlib.utils.twisted.sleep(unpause_time)
    sequence.unpause()

    yield deferred


@pytest_twisted.inlineCallbacks
def test_sequence_stop_while_paused(action_logger, assert_no_unhandled_errbacks):
    sequence = epyqlib.utils.twisted.Sequence()
    t = 0
    delay = 0.3
    for n in range(3):
        sequence.add_delayed(delay, action_logger.action, n)
        t += delay
        action_logger.expected.append(Result(time=t, data=n))

    pause_before = 2

    pause_time = action_logger.expected[pause_before].time - (delay / 2)

    action_logger.expected = action_logger.expected[:pause_before]

    deferred = sequence.run()

    yield epyqlib.utils.twisted.sleep(pause_time)
    sequence.pause()

    yield epyqlib.utils.twisted.sleep(1)
    sequence.cancel()

    with pytest.raises(twisted.internet.defer.CancelledError):
        yield deferred



#!/usr/bin/env python3

# TODO: """DocString if there is one"""

import can
from collections import OrderedDict
from epyqlib.abstractcolumns import AbstractColumns
import epyqlib.canneo
import functools
import json
import epyqlib.pyqabstractitemmodel
from epyqlib.treenode import TreeNode
from PyQt5.QtCore import Qt, QVariant, QModelIndex, pyqtSlot, QTimer
from PyQt5.QtWidgets import QFileDialog
import time

# See file COPYING in this source tree
__copyright__ = "Copyright 2016, EPC Power Corp."
__license__ = "GPLv2+"


class Columns(AbstractColumns):
    _members = ["name", "action"]


Columns.indexes = Columns.indexes()


class Node(TreeNode):
    def __init__(self, text, action=None):
        TreeNode.__init__(self)

        self.fields = Columns(name=text, action=action)

    # def to_ordered_dict(self):
    #     d = OrderedDict()
    #     for child in self.children:
    #         # TODO: actually store the action not its string
    #         d[child.fields.name] = child.fields.action
    #
    #     return d

    def from_ordered_dict(self, d):
        for child in self.children:
            value = d.get(child.fields.name, None)
            if value is not None:
                child.set_human_value(value)

    def find_child(self, value):
        for child in self.children:
            if child.fields.name == value or child.fields.action == value:
                return child

    @property
    def action(self):
        return self.fields.action

    @action.setter
    def action(self, action):
        self.fields.action = action

    def unique(self):
        return self


class ListMenuModel(epyqlib.pyqabstractitemmodel.PyQAbstractItemModel):
    def __init__(self, root, parent=None):
        epyqlib.pyqabstractitemmodel.PyQAbstractItemModel.__init__(
            self, root=root, alignment=Qt.AlignVCenter | Qt.AlignLeft, parent=parent
        )

        self.headers = Columns(name="Name", action="Action")

        self.set_root(root)

    @pyqtSlot(Node)
    def node_clicked(self, node):
        if node.fields.action is None:
            action = functools.partial(self.set_root, node)
        else:
            action = node.fields.action

        QTimer.singleShot(100, action)

    @pyqtSlot(bool)
    def esc_clicked(self, _):
        new_root = self.root.tree_parent
        if new_root is not None:
            self.node_clicked(new_root)


if __name__ == "__main__":
    import sys

    print("No script functionality here")
    sys.exit(1)  # non-zero is a failure



import logging
import can
import epyqlib.busproxy
import epyqlib.device
import epyqlib.twisted.busproxy
import epyqlib.twisted.cancalibrationprotocol as ccp
import pytest
import sys

from PyQt5.QtCore import QTimer

# import epyqlib.ticoff as ticoff
#
# files = [
#     '/epc/t/409/afe.out',
#     '/epc/t/409/ul1741.lc12_3.out',
#     '/epc/t/409/RS12_04c.out'
# ]
#
# files = {file: ticoff.Coff(file) for file in files}
#
# import itertools
#
#
# def endswap(l):
#     return itertools.chain(*((b, a) for a, b in zip(l[::2], l[1::2])))
#
#
# for file, c in sorted(files.items()):
#     print('\n\n\n----- {}'.format(file))
#     print('\n'.join([str((i,
#                           s.name,
#                           hex(s.virt_addr),
#                           s.virt_size,
#                           len(s.data),  # if s.data is not None else -1,
#                           'even' if len(s.data) % 2 else 'odd',
#                           ' '.join(list('{:02X}'.format(b) for b in
#                                         itertools.islice(endswap(s.data), 10))),
#                           # if s.data is not None else []
#                           ' '.join(list('{:02X}'.format(b) for b in
#                                         itertools.islice(endswap(s.data), max(0,
#                                                                               len(
#                                                                                   s.data) - 10),
#                                                          None))),
#                           # if s.data is not None else []
#                           s.data[:30]
#                           )) for i, s
#                      in enumerate(c.sections) if s.data is not None]))


@pytest.mark.require_device
def test_main():
    # TODO: CAMPid 03127876954165421679215396954697
    # https://github.com/kivy/kivy/issues/4182#issuecomment-253159955
    # fix for pyinstaller packages app to avoid ReactorAlreadyInstalledError
    if "twisted.internet.reactor" in sys.modules:
        del sys.modules["twisted.internet.reactor"]

    import qt5reactor

    qt5reactor.install()

    from twisted.internet import reactor

    real_bus = can.interface.Bus(bustype="socketcan", channel="can0")
    bus = epyqlib.busproxy.BusProxy(bus=real_bus)

    device = epyqlib.device.Device(
        file=epyqlib.tests.common.devices["customer"],
        node_id=247,
    )

    tx_signal = device.neo_frames.signal_by_path("CCP", "Connect", "CommandCounter")
    rx_signal = device.neo_frames.signal_by_path(
        "CCPResponse", "Connect", "CommandCounter"
    )
    protocol = ccp.Handler(
        endianness="little" if tx_signal.little_endian else "big",
        tx_id=tx_signal.frame.id,
        rx_id=rx_signal.frame.id,
    )
    transport = epyqlib.twisted.busproxy.BusProxy(
        protocol=protocol,
        reactor=reactor,
        bus=bus,
    )

    d = protocol.connect()
    d.addCallback(
        lambda _: protocol.set_mta(
            address_extension=ccp.AddressExtension.flash_memory,
            address=0x310000,
        )
    )
    d.addCallback(lambda _: protocol.disconnect())
    d.addBoth(logit)

    logging.debug("---------- started")
    QTimer.singleShot(3000, reactor.stop)
    reactor.run()


def logit(it):
    logging.debug("logit(): ({}) {}".format(type(it), it))


def test_identifier_type_error():
    with pytest.raises(ccp.IdentifierTypeError):
        ccp.HostCommand(code=ccp.CommandCode.connect, extended_id=False)


def test_payload_assignment_error():
    with pytest.raises(AttributeError):
        hc = ccp.HostCommand(code=ccp.CommandCode.connect)
        hc.payload = 0


def test_message_length_error():
    with pytest.raises(ccp.MessageLengthError):
        ccp.HostCommand(code=ccp.CommandCode.connect, dlc=5)



import pathlib

import attr
import pytest

import epyqlib.tests.common
import epyqlib.hildevice


@pytest.fixture
def factory_definition():
    original = pathlib.Path(epyqlib.tests.common.devices["factory"])
    with epyqlib.updateepc.updated(original) as updated:
        return epyqlib.hildevice.Definition.loadp(updated)


@pytest.mark.parametrize("version", [[1], [3]])
def test_definition_format_version_validator(factory_definition, version):
    with pytest.raises(epyqlib.hildevice.FormatVersionError):
        attr.evolve(factory_definition, format_version=version)


def test_definition_load(factory_definition):
    assert factory_definition.access_level_path is not None
    assert factory_definition.access_password_path is not None


def test_definition_loads():
    path = pathlib.Path(epyqlib.tests.common.devices["factory"])

    with epyqlib.updateepc.updated(path) as updated:
        with open(updated) as f:
            s = f.read()

        epyqlib.hildevice.Definition.loads(
            s=s,
            base_path=updated.parents[0],
        )


def test_load():
    device = epyqlib.hildevice.Device(
        definition_path=epyqlib.tests.common.devices["factory"],
    )
    device.load()


def test_device_reload():
    device = epyqlib.hildevice.Device(
        definition_path=epyqlib.tests.common.devices["factory"],
    )
    device.load()

    with pytest.raises(epyqlib.hildevice.AlreadyLoadedError):
        device.load()



#!/usr/bin/env python3

# TODO: """DocString if there is one"""

import epyqlib.utils.qt
import epyqlib.widgets.compactepc_ui
import epyqlib.widgets.epc


# See file COPYING in this source tree
__copyright__ = "Copyright 2016, EPC Power Corp."
__license__ = "GPLv2+"


class CompactEpc(epyqlib.widgets.epc.Epc):
    def __init__(self, parent=None, in_designer=False):
        super().__init__(
            parent=parent,
            ui_class=epyqlib.widgets.compactepc_ui.Ui_Form,
            in_designer=in_designer,
        )


if __name__ == "__main__":
    import sys

    print("No script functionality here")
    sys.exit(1)  # non-zero is a failure



import click

import epyqlib.cli.audit
import epyqlib.pm.valueset


@click.group()
def cli():
    pass


cli.add_command(epyqlib.pm.valueset.group)
cli.add_command(epyqlib.cli.audit.create_command(), name="audit")
try:
    import epyqlib.cli.phabricator_extract

    cli.add_command(
        epyqlib.cli.phabricator_extract.create_command(), name="phabricator_extract"
    )
except ModuleNotFoundError as mnfe:
    click.echo(
        f"Phabricator extract functionality is unavailable due to missing module: {mnfe.name}"
    )



import inspect
from typing import Callable, Coroutine

from PyQt5.QtCore import QObject
from PyQt5.QtWidgets import QDialog
from boto3_type_annotations.s3 import ServiceResource as S3Resource
from twisted.internet.defer import ensureDeferred

from epyqlib.tabs.files.cognito import CognitoHelper
from epyqlib.tabs.files.login_dialog import LoginDialog
from epyqlib.tabs.files.sync_config import SyncConfig, Vars

## Async function that takes a bool whether or not the user was just logged in
LoginListener = Callable[[bool], Coroutine]


class AwsLoginManager:
    _instance = None

    def __init__(self):
        if self._instance is not None:
            raise Exception(
                "Tried to create another instance of AwsLoginManager although one already exists."
            )
        self._listeners: [LoginListener] = []
        self._cognito_helper = CognitoHelper(SyncConfig.get_env())

    @staticmethod
    def get_instance():
        if AwsLoginManager._instance is None:
            AwsLoginManager._instance = AwsLoginManager()

        return AwsLoginManager._instance

    def is_logged_in(self) -> bool:
        return self._cognito_helper.is_user_logged_in()

    def show_login_window(self, parent: QObject = None):

        dialog: QDialog = LoginDialog(self)

        code = dialog.exec()

        if code == QDialog.Accepted:
            self._notify_listeners()

            # Enable auto-sync when the user logs in
            SyncConfig.get_instance().set(Vars.auto_sync, True)

    def authenticate(self, username: str, password: str):
        self._cognito_helper.authenticate(username, password)

    def log_user_out(self):
        self._cognito_helper.log_out()
        self._notify_listeners()

    def get_id_token(self) -> str:
        return self._cognito_helper._id_token

    def get_valid_id_token(self) -> str:
        if not self._cognito_helper.is_session_valid():
            self.refresh()

        return self.get_id_token()

    def get_s3_resource(self) -> S3Resource:
        return self._cognito_helper.get_s3_resource()

    def get_username(self):
        return self._cognito_helper.get_username()

    def get_user_customer(self):
        return self._cognito_helper.get_user_customer()

    def is_session_valid(self) -> bool:
        return self._cognito_helper.is_session_valid()

    def refresh(self, force=False):
        self._cognito_helper._refresh(force=force)

    ## Manage Listeners
    def _notify_listeners(self):
        # if login was successful, notify listeners
        for listener in self._listeners:
            result = listener(self._cognito_helper.is_user_logged_in())
            if inspect.iscoroutine(result):
                ensureDeferred(result)

    def register_listener(self, listener: LoginListener):
        self._listeners.append(listener)

    def remove_listener(self, listener: LoginListener):
        self._listeners.remove(listener)



import epyqlib.dualscale
import epyqlib.abstractpluginclass

# See file COPYING in this source tree
__copyright__ = "Copyright 2017, EPC Power Corp."
__license__ = "GPLv2+"


class DualScalePlugin(epyqlib.abstractpluginclass.AbstractPlugin):
    def __init__(self, parent=None):
        epyqlib.abstractpluginclass.AbstractPlugin.__init__(self, parent=parent)

        self._group = "EPC - Other"
        self._init = epyqlib.dualscale.DualScale
        self._module_path = "epyqlib.dualscale"
        self._name = "DualScale"



#!/usr/bin/env python3

# TODO: """DocString if there is one"""

import textwrap

from PyQt5.QtCore import pyqtProperty, pyqtSignal, QEvent, Qt
from PyQt5.QtGui import QMouseEvent
from PyQt5.QtWidgets import QMessageBox, QWidget

import epyqlib.widgets.abstractwidget


# See file COPYING in this source tree
__copyright__ = "Copyright 2016, EPC Power Corp."
__license__ = "GPLv2+"


class AbstractTxWidget(epyqlib.widgets.abstractwidget.AbstractWidget):
    edit = pyqtSignal(QWidget, QWidget)
    edited = pyqtSignal(float)

    def __init__(self, ui_class, parent=None, in_designer=False):
        self._tx = False

        super().__init__(
            ui_class=ui_class,
            parent=parent,
            in_designer=in_designer,
        )

        for widget in self.findChildren(QWidget):
            if widget.property("editable_click"):
                widget.installEventFilter(self)

        self._period = None
        self.check_range = True

    @pyqtProperty(bool)
    def tx(self):
        return self._tx

    @tx.setter
    def tx(self, tx):
        self._tx = bool(tx)

        self.set_signal(signal=self.signal_object)
        if self.ui is not None:
            self.ui.value.setDisabled(not self.tx)

    def meta_set_value(self, value):
        if not self.ui.value.hasFocus():
            self.set_value(value)

    def user_set_value(self, value):
        self.signal_object.set_human_value(value, check_range=True)
        self.edited.emit(value)

    def eventFilter(self, qobject, qevent):
        if (
            isinstance(qevent, QMouseEvent)
            and self.tx
            and qevent.button() == Qt.LeftButton
            and qevent.type() == QEvent.MouseButtonRelease
            and qobject.rect().contains(qevent.localPos().toPoint())
        ):
            self.edit.emit(self, self)

            return True

        return False

    def set_signal(self, signal=None, force_update=False):
        if signal is not None and self.tx:
            signal.frame.user_send_control = False
            period = signal.frame.cycle_time
            if period is None:
                self._period = None
            else:
                self._period = float(period) / 1000
        else:
            self._period = None

        epyqlib.widgets.abstractwidget.AbstractWidget.set_signal(
            self, signal, force_update=force_update
        )

    def update_connection(self, signal=None):
        if signal is not self.signal_object:
            if self.signal_object is not None:
                self.signal_object.frame.cyclic_request(self, None)

            if signal is not None and self.tx and self._period is not None:
                signal.frame.cyclic_request(self, self._period)

        epyqlib.widgets.abstractwidget.AbstractWidget.update_connection(self, signal)

    def widget_value_changed(self, value):
        if self.signal_object is not None and self.tx:
            try:
                self.signal_object.set_human_value(
                    value,
                    check_range=self.check_range,
                )
            except epyqlib.canneo.OutOfRangeError as e:
                message = textwrap.dedent(
                    """\
                Frame: {frame}
                Signal: {signal}

                Error: {error}
                """
                ).format(
                    frame=self.signal_object.frame.name,
                    signal=self.signal_object.name,
                    error=str(e),
                )

                epyqlib.utils.qt.dialog(
                    parent=self,
                    message=message,
                    icon=QMessageBox.Critical,
                )
            else:
                if self._period is None:
                    self.signal_object.frame.send_now(
                        signals=(self.signal_object,),
                    )

    def set_check_range(self, check):
        self.check_range = check


if __name__ == "__main__":
    import sys

    print("No script functionality here")
    sys.exit(1)  # non-zero is a failure



import logging

import pytest

import epyqlib.busproxy
import epyqlib.device
import epyqlib.twisted.busproxy
import epyqlib.tests.common


def test_range_override_hidden(qtbot):
    device = epyqlib.device.Device(
        file=epyqlib.tests.common.devices["customer"],
        node_id=247,
    )

    children = device.ui.findChildren(epyqlib.nvview.NvView)

    assert len(children) > 0

    for view in children:
        checkbox = view.ui.enforce_range_limits_check_box
        assert not checkbox.isVisibleTo(checkbox.parent())


@pytest.mark.factory
def test_range_override_visible(qtbot):
    device = epyqlib.device.Device(
        file=epyqlib.tests.common.devices["factory"],
        node_id=247,
    )

    children = device.ui.findChildren(epyqlib.nvview.NvView)

    assert len(children) > 0

    for view in children:
        checkbox = view.ui.enforce_range_limits_check_box
        assert checkbox.isVisibleTo(checkbox.parent())


@pytest.mark.factory
def test_secret_masked(qtbot):
    secret_mask = "<secret>"

    device = epyqlib.device.Device(
        file=epyqlib.tests.common.devices["factory"],
        node_id=247,
    )

    secret_nv = tuple(nv for nv in device.nvs.all_nv() if nv.secret)

    assert len(secret_nv) > 0
    for nv in secret_nv:
        nv.set_meta("1234", epyqlib.nv.MetaEnum.user_default)
        assert nv.fields.user_default == secret_mask
        nv.set_meta("1234", epyqlib.nv.MetaEnum.factory_default)
        assert nv.fields.factory_default == secret_mask
        nv.set_meta("1234", epyqlib.nv.MetaEnum.value)
        assert nv.fields.value == secret_mask


def logit(it):
    logging.debug("logit(): ({}) {}".format(type(it), it))



from PyQt5 import QtWidgets

import epyqlib.variableselectionview_ui
import epyqlib.utils.qt


# See file COPYING in this source tree
__copyright__ = "Copyright 2016, EPC Power Corp."
__license__ = "GPLv2+"


class VariableSelectionView(QtWidgets.QWidget):
    def __init__(self, parent=None, in_designer=False):
        super().__init__(parent=parent)

        self.in_designer = in_designer

        self.ui = epyqlib.variableselectionview_ui.Ui_Form()
        self.ui.setupUi(self)

        self.ui.searchbox.connect_to_view(
            view=self.ui.tree_view,
            column=epyqlib.variableselectionmodel.Columns.indexes.name,
        )

    def set_model(self, model):
        self.ui.tree_view.setModel(model)
        model.setSortRole(epyqlib.utils.qt.UserRoles.sort)

        header = self.ui.tree_view.header()
        header.setStretchLastSection(False)
        header.setSectionResizeMode(header.ResizeToContents)

    def set_sorting_enabled(self, enabled):
        self.ui.tree_view.setSortingEnabled(enabled)

    def sort_by_column(self, column, order):
        self.ui.tree_view.sortByColumn(column, order)

    @property
    def model(self):
        return self.ui.tree_view.model()



import collections
import json
import os
import pathlib
import time
import zipfile

import attr
import can
import dotenv
import pytest
import pytest_twisted
import twisted

import epyqlib.autodevice.build
import epyqlib.busproxy
import epyqlib.device
import epyqlib.tests.common
import epyqlib.utils.general

example_archive_code = "the archive code"
example_access_password = "1"
example_access_level = "2"


@attr.s
class AutoDevice:
    version = attr.ib()
    path = attr.ib()


def create_example_auto_device(
    version,
    target,
    access_password,
    archive_code=example_archive_code,
    access_level=example_access_level,
    serial_number=None,
    parameter_type="pmvs",
    template=epyqlib.autodevice.build.raw_template / "auto_parameters.epc",
):
    builder = epyqlib.autodevice.build.Builder()
    builder.archive_code = archive_code

    if serial_number is not None:
        builder.required_serial_number = serial_number

    builder.set_access_level_names(
        password_name="AccessLevel:Password",
        access_level_name="AccessLevel:Level",
    )

    for access_input in builder.access_parameters:
        if "Password" in access_input.node:
            access_input.value = access_password
        elif "Level" in access_input.node:
            access_input.value = access_level

    builder.set_template(
        path=template,
        archive=template.suffix.casefold() == ".zip",
    )

    device_files = epyqlib.tests.common.new_devices()[(version, "factory")]

    with open(device_files.device) as f:
        original_raw_dict = json.load(
            f,
            object_pairs_hook=collections.OrderedDict,
        )

    builder.set_original_raw_dict(original_raw_dict)

    if parameter_type == "pmvs":
        builder.load_pmvs(device_files.pmvs)
    elif parameter_type == "epp":
        builder.load_epp_paths(
            parameter_path=device_files.epp,
            can_path=device_files.can,
        )

    builder.set_target(path=target)

    with open(device_files.can, "rb") as f:
        can_contents = f.read()

    builder.create(can_contents=can_contents)

    return AutoDevice(version=version, path=target)


@pytest.fixture(scope="session")
def access_password():
    env = dict(os.environ)
    dotenv_normal = dotenv.dotenv_values()
    dotenv_local = dotenv.dotenv_values(dotenv.find_dotenv(".env.local"))

    env.update(dotenv_normal)
    env.update(dotenv_local)

    return env.get("ST_ACCESS_PASSWORD")


@pytest.fixture(params=["epp", "pmvs"], scope="session")
def parameter_type(request):
    return request.param


versions = ["develop", "v1.2.5"]
# version_to_test = 'v1.2.5'
version_to_test = "develop"


@pytest.fixture(params=[version_to_test], scope="session")
def version(request):
    return request.param


@pytest.fixture(scope="session")
def auto_device(qapp, version, parameter_type, access_password, tmpdir_factory):
    target = tmpdir_factory.mktemp("auto_device").join("auto_device.epz")

    yield create_example_auto_device(
        version=version,
        target=target,
        parameter_type=parameter_type,
        access_password=access_password,
    )


@pytest.fixture()
def bus():
    real_bus = can.interface.Bus(bustype="socketcan", channel="can0")
    bus = epyqlib.busproxy.BusProxy(bus=real_bus)

    yield bus

    bus.terminate()
    real_bus.shutdown()


@twisted.internet.defer.inlineCallbacks
def wait_until_present(device, period=1):
    start = time.monotonic()

    yield epyqlib.utils.twisted.sleep(2 * device.connection_monitor.timeout)

    while True:
        yield epyqlib.utils.twisted.sleep(period)

        if device.connection_monitor.present:
            break

    save_time = time.monotonic() - start
    print(f"Waited {save_time:0.2f} seconds for EE to save")


def test_contains_epc(auto_device):
    with zipfile.ZipFile(auto_device.path) as z:
        assert any(name.endswith(".epc") for name in z.namelist())


def test_from_zip(version, parameter_type, access_password, tmpdir):
    temporary_directory = pathlib.Path(tmpdir)
    target = temporary_directory / "auto_device.epz"

    zip_path = temporary_directory / "archive.zip"

    epyqlib.autodevice.build.create_template_archive(zip_path)

    auto_device = create_example_auto_device(
        version=version,
        target=target,
        parameter_type=parameter_type,
        access_password=access_password,
        template=zip_path,
    )

    with zipfile.ZipFile(auto_device.path) as z:
        assert any(name.endswith(".epc") for name in z.namelist())


@pytest.mark.require_device
@pytest_twisted.inlineCallbacks
def test_general_load(qapp, auto_device):
    device = epyqlib.device.Device(
        file=auto_device.path,
        archive_code=example_archive_code,
        node_id=247,
        bus=epyqlib.busproxy.BusProxy(),
    )

    with device.bus.managed_real_bus(bustype="socketcan", channel="can0"):
        with device:
            yield device.extension.no_gui_load_parameters()
            yield wait_until_present(device)


@pytest.mark.require_device
@pytest_twisted.inlineCallbacks
def test_invalid_serial(qapp, auto_device, access_password):
    # string should never match a serial number
    create_example_auto_device(
        version=auto_device.version,
        target=auto_device.path,
        serial_number="a",
        access_password=access_password,
    )

    device = epyqlib.device.Device(
        file=auto_device.path,
        archive_code=example_archive_code,
        node_id=247,
        bus=epyqlib.busproxy.BusProxy(),
    )

    with device.bus.managed_real_bus(bustype="socketcan", channel="can0"):
        with device:
            with pytest.raises(
                epyqlib.utils.general.UnmatchedSerialNumberError,
            ):
                yield device.extension.no_gui_load_parameters()
            yield wait_until_present(device)


@pytest.mark.require_device
@pytest_twisted.inlineCallbacks
def test_valid_serial(qapp, auto_device, tmpdir, access_password):
    temporary_directory = pathlib.Path(tmpdir)

    device = epyqlib.device.Device(
        file=auto_device.path,
        archive_code=example_archive_code,
        node_id=247,
        bus=epyqlib.busproxy.BusProxy(),
    )

    with device.bus.managed_real_bus(bustype="socketcan", channel="can0"):
        with device:
            serial_nv = device.extension.nvs.signal_from_names(
                *device.extension.serial_number_names,
            )
            present_serial, _ = yield device.extension.nv_protocol.read(
                nv_signal=serial_nv,
                meta=epyqlib.nv.MetaEnum.value,
            )

    device_path = temporary_directory / "serial_locked_auto_device.epz"

    auto_device = create_example_auto_device(
        version=auto_device.version,
        target=device_path,
        serial_number=present_serial,
        access_password=access_password,
    )

    device = epyqlib.device.Device(
        file=auto_device.path,
        archive_code=example_archive_code,
        node_id=247,
        bus=epyqlib.busproxy.BusProxy(),
    )

    with device.bus.managed_real_bus(bustype="socketcan", channel="can0"):
        with device:
            yield device.extension.no_gui_load_parameters()
            yield wait_until_present(device)



#!/usr/bin/env python3

# TODO: get some docstrings in here!

import logging

logger = logging.getLogger(__name__)

import attr
import can
import canmatrix.formats
import collections
import copy
import decimal
import epyqlib.canneo
import epyqlib.deviceextension
import epyqlib.faultlogmodel

# https://www.riverbankcomputing.com/pipermail/pyqt/2018-December/041218.html
import epyqlib.tabs.files.filesview

try:
    import epyqlib.resources.code
except ImportError:
    pass  # we will catch the failure to open the file
import epyqlib.nv
import epyqlib.nvview
import epyqlib.overlaylabel
import epyqlib.scripting
import epyqlib.twisted.loopingset
import epyqlib.txrx
import epyqlib.txrxview
import epyqlib.utils.qt
import epyqlib.utils.j1939
import epyqlib.variableselectionmodel
import functools
import importlib.util
import io
import itertools
import json
import math
import os
import pathlib
import shutil
import tempfile
import textwrap
import twisted.internet.task
import uuid
import zipfile
from twisted.internet.defer import setDebugging

setDebugging(True)

from collections import OrderedDict
from enum import Enum, unique
from epyqlib.busproxy import BusProxy
import epyqlib.updateepc
from epyqlib.widgets.abstractwidget import AbstractWidget
from PyQt5 import uic
from PyQt5.QtCore import (
    pyqtSlot,
    Qt,
    QFile,
    QFileInfo,
    QTextStream,
    QObject,
    QSortFilterProxyModel,
    QIODevice,
    QTimer,
)
from PyQt5.QtWidgets import (
    QWidget,
    QMessageBox,
    QInputDialog,
    QLineEdit,
    QVBoxLayout,
    QStackedLayout,
)
from PyQt5 import QtCore

# See file COPYING in this source tree
__copyright__ = "Copyright 2018, EPC Power Corp."
__license__ = "GPLv2+"


class CancelError(Exception):
    pass


@unique
class Elements(Enum):
    dash = 1
    tx = 2
    rx = 3
    variables = 4
    nv = 5
    scripting = 6
    fault_log = 7
    files = 8
    information = 9


@unique
class Tabs(Enum):
    dashes = 1
    txrx = 2
    variables = 3
    nv = 4
    scripting = 5
    fault_log = 6
    files = 7
    information = 8

    @classmethod
    def defaults(cls):
        return set(cls) - {
            cls.variables,
            cls.scripting,
            cls.fault_log,
            cls.files,
        }


def j1939_node_id_adjust(message_id, device_id, to_device, controller_id):
    # CCP stuff is in bounds now but leave this for backwards compatibility
    if message_id > 0x1FFFFF00:
        return message_id

    id = epyqlib.utils.j1939.Id.unpack(message_id)

    if to_device:
        source = controller_id
        destination = device_id
    else:
        source = device_id
        destination = controller_id

    if id.is_pdu1():
        if destination is not None:
            id.destination_address = destination

    if source is not None:
        id.source_address = source

    return id.pack()


def simple_node_id_adjust(message_id, device_id, to_device, controller_id):
    return message_id + device_id


node_id_types = OrderedDict(
    [("j1939", j1939_node_id_adjust), ("simple", simple_node_id_adjust)]
)


@attr.s
class CanConfiguration:
    data_logger_reset_signal_path = attr.ib()
    data_logger_recording_signal_path = attr.ib()
    data_logger_configuration_is_valid_signal_path = attr.ib()
    monitor_frame = attr.ib()


can_configurations = {
    "original": CanConfiguration(
        data_logger_reset_signal_path=("CommandModeControl", "ResetDatalogger"),
        data_logger_recording_signal_path=("StatusBits", "DataloggerRecording"),
        data_logger_configuration_is_valid_signal_path=(
            "StatusBits",
            "DataloggerConfigurationIsValid",
        ),
        monitor_frame="StatusBits",
    ),
    "j1939": CanConfiguration(
        data_logger_reset_signal_path=(
            "ParameterQuery",
            "DataloggerConfig",
            "ResetDatalogger",
        ),
        data_logger_recording_signal_path=(
            "ParameterQuery",
            "DataloggerStatus",
            "DataloggerRecording",
        ),
        data_logger_configuration_is_valid_signal_path=(
            "ParameterQuery",
            "DataloggerStatus",
            "DataloggerConfigurationIsValid",
        ),
        monitor_frame="StatusBits",
    ),
}


def load(file):
    if isinstance(file, str):
        pass
    elif isinstance(file, io.IOBase):
        pass


def ignore_timeout(failure):
    acceptable_errors = (
        epyqlib.twisted.nvs.RequestTimeoutError,
        epyqlib.twisted.nvs.SendFailedError,
        epyqlib.twisted.nvs.CanceledError,
    )
    if failure.type in acceptable_errors:
        return None

    return epyqlib.utils.twisted.errbackhook(failure)


def load_matrix(path):
    matrix = list(canmatrix.formats.loadp(path).values())[0]

    if hasattr(matrix, "load_errors"):
        # https://github.com/ebroecker/canmatrix/pull/199
        if len(matrix.load_errors) > 0:
            first_error = matrix.load_errors[0]
            raise Exception(
                f"{type(first_error).__name__}: {first_error}",
            ) from first_error

    return matrix


class Device:
    def __init__(self, *args, **kwargs):
        self.bus = None
        self.from_zip = False
        self.device_sha = "N/A"
        self.device_file = "N/A"

        if kwargs.get("file", None) is not None:
            constructor = self._init_from_file
        else:
            constructor = self._init_from_parameters

        constructor(*args, **kwargs)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.terminate()

    def terminate(self):
        if self.bus is not None:
            while self.notifiees:
                self.bus.notifier.discard(self.notifiees.pop())

        self.bus = None

        self.neo_frames.terminate()
        try:
            self.ui.tabs.currentChanged.disconnect()
        except TypeError:
            # We don't really mind if there aren't any slots connect
            pass
        if self.nv_looping_set is not None:
            self.nv_looping_set.stop()
        if self.nv_tab_looping_set is not None:
            self.nv_tab_looping_set.stop()
        if self.nvs is not None:
            self.nvs.terminate()
        if self.widget_nvs is not None:
            self.widget_nvs.terminate()
        if self.nv_views is not None:
            for view in self.nv_views:
                view.terminate()

        terminate_extension = getattr(self.extension, "terminate", None)
        if terminate_extension is not None:
            terminate_extension()
        self.extension.device = None

        self.connection_monitor.terminate()
        self.connection_monitor = None

        self.dash_uis = None
        self.loaded_uis = None
        self.neo_frames = None
        self.ui = None
        # TODO: why does this contain something other than paths :[
        self.ui_paths = None
        self.uis = None
        self.widget_nv_frames = None
        self.widget_nvs = None

        logging.debug("{} terminated".format(object.__repr__(self)))

    def __del__(self):
        if self.bus is not None:
            self.bus.set_bus()

    def _init_from_file(self, file, only_for_files=False, **kwargs):
        self.device_file = file
        extension = os.path.splitext(file)[1].casefold()

        if extension in (".epz", ".zip"):
            zip_file = zipfile.ZipFile(file)
            self._init_from_zip(zip_file, **kwargs)
        else:
            try:
                self.config_path = os.path.abspath(file)
                file = open(file, "r")
            except TypeError:
                return
            else:
                converted_directory = None
                final_file = file
                if not epyqlib.updateepc.is_latest(file.name):
                    converted_directory = tempfile.TemporaryDirectory()
                    final_file = open(
                        epyqlib.updateepc.convert(
                            file.name,
                            converted_directory.name,
                        )
                    )
                self.config_path = os.path.abspath(final_file.name)

                self._load_config(
                    file=final_file, only_for_files=only_for_files, **kwargs
                )

                if final_file is not file:
                    final_file.close()

                if converted_directory is not None:
                    converted_directory.cleanup()

    def _load_config(
        self,
        file,
        elements=None,
        tabs=None,
        rx_interval=0,
        edit_actions=None,
        only_for_files=False,
        node_id=None,
        **kwargs,
    ):
        if tabs is None:
            tabs = Tabs.defaults()

        self.on_offline_bus = False

        self.node_id = node_id

        self.elements = Elements if elements == None else elements
        self.elements = set(Elements)

        s = file.read()
        d = json.loads(s, object_pairs_hook=OrderedDict)
        self.raw_dict = d
        d.setdefault("nv_meta_enum", None)
        d.setdefault("access_level_path", None)
        d.setdefault(
            "access_password_path",
            "ParameterQuery;FactoryAccess;FactoryAccess",
        )

        self.module_path = d.get("module", None)
        self.plugin = None
        if self.module_path is None:
            module = epyqlib.deviceextension
        else:
            spec = importlib.util.spec_from_file_location(
                "extension", self.absolute_path(self.module_path)
            )
            module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(module)

        extension_class = module.DeviceExtension

        import weakref

        self.extension = extension_class(device=weakref.ref(self))

        path = os.path.dirname(file.name)
        for ui_path_name in ["ui_path", "ui_paths", "menu"]:
            try:
                json_ui_paths = d[ui_path_name]
                break
            except KeyError:
                pass
        else:
            json_ui_paths = {}

        if not isinstance(json_ui_paths, dict):
            json_ui_paths = {"Dash": json_ui_paths}

        self.ui_paths = json_ui_paths

        hierarchy_path = d.get("parameter_hierarchy", None)
        if hierarchy_path is None:
            hierarchy = None
        else:
            with open(self.absolute_path(hierarchy_path)) as f:
                hierarchy = json.load(f)

        for tab in Tabs:
            try:
                value = d["tabs"][tab.name]
            except KeyError:
                pass
            else:
                if int(value):
                    tabs.add(tab)
                else:
                    tabs.discard(tab)

        depended_on = set()

        if Tabs.txrx not in tabs:
            self.elements.discard(Elements.tx)
            self.elements.discard(Elements.rx)

        if Tabs.variables not in tabs:
            self.elements.discard(Elements.variables)

        if Tabs.nv not in tabs:
            self.elements.discard(Elements.nv)

        if Tabs.scripting not in tabs:
            self.elements.discard(Elements.scripting)

        if Tabs.fault_log not in tabs:
            self.elements.discard(Elements.fault_log)
        else:
            depended_on.add(Elements.nv)

        if Tabs.files not in tabs:
            self.elements.discard(Elements.files)

        if Tabs.information not in tabs:
            self.elements.discard(Elements.information)

        self.elements |= depended_on

        self.referenced_files = [
            f
            for f in [
                d.get("module", None),
                d.get("can_path", None),
                d.get("compatibility", None),
                d.get("parameter_defaults", None),
                d.get("parameter_hierarchy", None),
                *self.ui_paths.values(),
                *module.referenced_files(self.raw_dict),
            ]
            if f is not None
        ]

        self.compatibility_shas = []
        compatibility_file = d.get("compatibility", None)
        if compatibility_file is not None:
            compatibility_file = os.path.join(
                os.path.dirname(self.config_path), compatibility_file
            )
            with open(compatibility_file) as file:
                s = file.read()
                c = json.loads(s, object_pairs_hook=OrderedDict)

            self.compatibility_shas.extend(c.get("shas", []))

        if not only_for_files:
            self.can_path = os.path.join(path, d["can_path"])
            with open(self.can_path, "rb") as f:
                self.can_contents = f.read()

            self.node_id_type = d.get("node_id_type", next(iter(node_id_types))).lower()
            if self.node_id is None:
                self.node_id = d.get("node_id")
            if self.node_id is None and self.node_id_type == "j1939":
                self.node_id, ok = QInputDialog.getInt(
                    None,
                    *(("Converter Node ID",) * 2),
                    247,
                    0,
                    247,
                )

                if not ok:
                    raise CancelError("User canceled node ID dialog")
            self.node_id = int(self.node_id)
            self.controller_id = int(d.get("controller_id", 65))
            self.node_id_adjust = functools.partial(
                node_id_types[self.node_id_type],
                device_id=self.node_id,
                controller_id=self.controller_id,
            )

            self._init_from_parameters(
                uis=self.ui_paths,
                serial_number=d.get("serial_number", ""),
                name=d.get("name", ""),
                tabs=tabs,
                rx_interval=rx_interval,
                edit_actions=edit_actions,
                nv_configuration=d.get("nv_configuration"),
                can_configuration=d.get("can_configuration"),
                hierarchy=hierarchy,
                **kwargs,
            )

    def _init_from_zip(
        self,
        zip_file,
        rx_interval=0,
        archive_code=None,
        **kwargs,
    ):
        self.from_zip = True
        path = tempfile.mkdtemp()

        code = epyqlib.utils.qt.get_code()

        if archive_code is not None:
            code = archive_code.encode("ascii")
            zip_file.extractall(path=path, pwd=code)
        else:
            while True:
                try:
                    zip_file.extractall(path=path, pwd=code)
                except RuntimeError:
                    code, ok = QInputDialog.getText(
                        None, ".epz Password", ".epz Password", QLineEdit.Password
                    )

                    if not ok:
                        raise CancelError("User canceled password dialog")

                    code = code.encode("ascii")
                else:
                    break

        # TODO error dialog if no .epc found in zip file
        filename_epc = None
        filename_sha = None
        for directory, directories, files in os.walk(path):
            for f in files:
                logger.info(f)
                if os.path.splitext(f)[1] == ".epc":
                    filename_epc = os.path.join(path, directory, f)
                elif f == "sha":
                    filename_sha = os.path.join(path, directory, f)

            if filename_epc is not None and filename_sha is not None:
                break

        if filename_sha is not None:
            with open(filename_sha, "r") as f:
                self.device_sha = f.readline().strip()

        self.config_path = os.path.abspath(filename_epc)

        converted_directory = None
        if not epyqlib.updateepc.is_latest(filename_epc):
            converted_directory = tempfile.TemporaryDirectory()
            file = epyqlib.updateepc.convert(filename_epc, converted_directory.name)
            self.config_path = os.path.abspath(file)

        with open(filename_epc, "r") as f:
            self._load_config(f, rx_interval=rx_interval, **kwargs)

        if converted_directory is not None:
            converted_directory.cleanup()
        shutil.rmtree(path)

    def traverse(self, dict_node):
        for key, value in dict_node.items():
            if isinstance(value, dict):
                self.traverse(value)
            elif value.endswith(".ui"):
                path = value
                try:
                    dict_node[key] = self.loaded_uis[path]
                except KeyError:
                    # TODO: CAMPid 9549757292917394095482739548437597676742
                    if not QFileInfo(path).isAbsolute():
                        ui_file = os.path.join(
                            QFileInfo.absolutePath(QFileInfo(self.config_path)), path
                        )
                    else:
                        ui_file = path
                    ui_file = QFile(ui_file)
                    ui_file.open(QFile.ReadOnly | QFile.Text)
                    ts = QTextStream(ui_file)
                    sio = io.StringIO(ts.readAll())
                    dict_node[key] = uic.loadUi(sio)
                    dict_node[key].file_name = path
                    self.loaded_uis[path] = dict_node[key]

    def _init_from_parameters(
        self,
        uis,
        serial_number,
        name,
        bus=None,
        tabs=None,
        rx_interval=0,
        edit_actions=None,
        nv_configuration=None,
        can_configuration=None,
        hierarchy=None,
    ):
        if tabs is None:
            tabs = Tabs.defaults()

        if can_configuration is None:
            can_configuration = "original"

        can_configuration = can_configurations[can_configuration]

        self.nvs = None
        self.widget_nvs = None

        self.bus_online = False
        self.bus_tx = False

        self.bus = BusProxy(bus=bus)

        self.nv_looping_set = None
        self.nv_tab_looping_set = None

        self.rx_interval = rx_interval
        self.serial_number = serial_number
        self.name = "{name} :{id}".format(name=name, id=self.node_id)
        self.nickname = ""
        self.auto_read_nv_widget_min_max = True

        self.nv_views = None

        device_ui = "device.ui"
        # TODO: CAMPid 9549757292917394095482739548437597676742
        if not QFileInfo(device_ui).isAbsolute():
            ui_file = os.path.join(
                QFileInfo.absolutePath(QFileInfo(__file__)), device_ui
            )
        else:
            ui_file = device_ui
        ui_file = QFile(ui_file)
        ui_file.open(QFile.ReadOnly | QFile.Text)
        ts = QTextStream(ui_file)
        sio = io.StringIO(ts.readAll())
        self.ui = uic.loadUi(sio)
        self.loaded_uis = {}

        self.traverse(uis)

        # TODO: yuck, actually tidy the code
        self.dash_uis = uis

        notifiees = []

        # Load the matrix once and use as needed as a speed optimization.
        matrix = load_matrix(self.can_path)

        if Elements.dash in self.elements:
            self.uis = self.dash_uis

            # TODO: this is icky
            if Elements.tx not in self.elements:
                self.neo_frames = epyqlib.canneo.Neo(
                    matrix=matrix, bus=self.bus, rx_interval=self.rx_interval
                )

                notifiees.append(self.neo_frames)

        if Elements.rx in self.elements:
            # TODO: the repetition here is not so pretty
            neo_rx = epyqlib.canneo.Neo(
                matrix=matrix,
                frame_class=epyqlib.txrx.MessageNode,
                signal_class=epyqlib.txrx.SignalNode,
                node_id_adjust=self.node_id_adjust,
                strip_summary=False,
            )

            rx = epyqlib.txrx.TxRx(tx=False, neo=neo_rx)
            notifiees.append(rx)
            rx_model = epyqlib.txrx.TxRxModel(rx)

            # TODO: put this all in the model...
            rx.changed.connect(rx_model.changed)
            rx.begin_insert_rows.connect(rx_model.begin_insert_rows)
            rx.end_insert_rows.connect(rx_model.end_insert_rows)

        if Elements.tx in self.elements:
            message_node_tx_partial = functools.partial(
                epyqlib.txrx.MessageNode, tx=True
            )
            signal_node_tx_partial = functools.partial(epyqlib.txrx.SignalNode, tx=True)
            neo_tx = epyqlib.canneo.Neo(
                matrix=matrix,
                frame_class=message_node_tx_partial,
                signal_class=signal_node_tx_partial,
                node_id_adjust=self.node_id_adjust,
            )
            notifiees.extend(f for f in neo_tx.frames if f.mux_name is None)

            self.neo_frames = neo_tx

            tx = epyqlib.txrx.TxRx(tx=True, neo=neo_tx, bus=self.bus)
            tx_model = epyqlib.txrx.TxRxModel(tx)
            tx.changed.connect(tx_model.changed)

        # TODO: something with sets instead?
        if Elements.rx in self.elements or Elements.tx in self.elements:
            txrx_views = self.ui.findChildren(epyqlib.txrxview.TxRxView)
            if len(txrx_views) > 0:
                # TODO: actually find them and actually support multiple
                pairs = (
                    (self.ui.rx, rx_model),
                    (self.ui.tx, tx_model),
                )
                column = epyqlib.txrx.Columns.indexes.name
                for view, model in pairs:
                    if model.root.tx:
                        proxy = epyqlib.utils.qt.PySortFilterProxyModel(
                            filter_column=column,
                        )
                        proxy.setSortCaseSensitivity(Qt.CaseInsensitive)
                        proxy.setSourceModel(model)
                        view.setModel(proxy)
                        view.set_sorting_enabled(True)
                        view.sort_by_column(column=column, order=Qt.AscendingOrder)
                    else:
                        view.setModel(model)

        self.widget_nvs = None
        if Elements.nv in self.elements:
            self.frames_nv = epyqlib.canneo.Neo(
                matrix=matrix,
                frame_class=epyqlib.nv.Frame,
                signal_class=epyqlib.nv.Nv,
                node_id_adjust=self.node_id_adjust,
                strip_summary=False,
            )

            self.nv_looping_set = epyqlib.twisted.loopingset.Set()
            self.nv_tab_looping_set = epyqlib.twisted.loopingset.Set()

            access_level_path = self.raw_dict["access_level_path"]
            if access_level_path is not None:
                access_level_path = access_level_path.split(";")

            access_password_path = self.raw_dict["access_password_path"]
            if access_password_path is not None:
                access_password_path = access_password_path.split(";")

            def none_or_uuid(uuid_string):
                if uuid_string is None:
                    return None

                return uuid.UUID(uuid_string)

            parameter_uuids = self.raw_dict.get("parameter_uuids", {})
            serial_number_uuid = none_or_uuid(
                parameter_uuids.get("serial_number"),
            )

            # TODO: CAMPid 0794311304143707516085683164039671793972
            if self.raw_dict["nv_meta_enum"] == "Meta":
                self.metas = epyqlib.nv.meta_limits_first
            else:
                self.metas = (epyqlib.nv.MetaEnum.value,)

            self.nvs = epyqlib.nv.Nvs(
                neo=self.frames_nv,
                bus=self.bus,
                configuration=nv_configuration,
                hierarchy=hierarchy,
                metas=self.metas,
                access_level_path=access_level_path,
                access_password_path=access_password_path,
                serial_number_uuid=serial_number_uuid,
            )

            default_metas = [
                meta
                for meta in (
                    epyqlib.nv.MetaEnum.user_default,
                    epyqlib.nv.MetaEnum.factory_default,
                )
                if meta not in self.metas
            ]

            if len(default_metas) > 0 and "parameter_defaults" in self.raw_dict:
                parameter_defaults_path = os.path.join(
                    os.path.dirname(self.config_path),
                    self.raw_dict["parameter_defaults"],
                )
                with open(parameter_defaults_path) as f:
                    self.nvs.defaults_from_dict(
                        d=json.load(f, parse_float=decimal.Decimal),
                        default_metas=default_metas,
                    )

            for nv in self.nvs.all_nv():
                if isinstance(nv, epyqlib.nv.Nv):
                    if epyqlib.nv.MetaEnum.minimum not in self.metas:
                        if nv.min is not None:
                            nv.set_meta(
                                data=nv.min,
                                meta=epyqlib.nv.MetaEnum.minimum,
                            )

                    if epyqlib.nv.MetaEnum.maximum not in self.metas:
                        if nv.max is not None:
                            nv.set_meta(
                                data=nv.max,
                                meta=epyqlib.nv.MetaEnum.maximum,
                            )

            self.widget_frames_nv = epyqlib.canneo.Neo(
                matrix=matrix,
                frame_class=epyqlib.nv.Frame,
                signal_class=epyqlib.nv.Nv,
                node_id_adjust=self.node_id_adjust,
            )
            self.widget_nvs = epyqlib.nv.Nvs(
                neo=self.widget_frames_nv,
                bus=self.bus,
                stop_cyclic=self.nv_looping_set.stop,
                start_cyclic=self.nv_looping_set.start,
                configuration=nv_configuration,
                serial_number_uuid=serial_number_uuid,
            )
            notifiees.append(self.widget_nvs)

            self.nv_views = self.ui.findChildren(epyqlib.nvview.NvView)
            if len(self.nv_views) > 0:
                nv_model = epyqlib.nv.NvModel(self.nvs)
                self.nvs.changed.connect(nv_model.changed)

                self.first_nv_view = self.nv_views[0]

                column = epyqlib.nv.Columns.indexes.name
                for view in self.nv_views:
                    view.set_device(self)
                    view.set_can_contents(
                        can_contents=self.can_contents,
                        suffix=pathlib.Path(self.can_path).suffix,
                    )
                    if self.nvs.access_level_node is not None:
                        view.set_access_level_signal_path(
                            path=self.nvs.access_level_node.signal_path(),
                        )

                    sort_proxy = epyqlib.utils.qt.PySortFilterProxyModel(
                        filter_column=column,
                    )
                    sort_proxy.setSortCaseSensitivity(Qt.CaseInsensitive)
                    sort_proxy.setSourceModel(nv_model)

                    diff_proxy = epyqlib.utils.qt.DiffProxyModel(
                        columns=epyqlib.nv.diffable_columns,
                        reference_column=(epyqlib.nv.Columns.indexes.user_default),
                        diff_highlights={
                            QtCore.Qt.ItemDataRole.BackgroundRole: epyqlib.nv.diff_highlight,
                        },
                        reference_highlights={
                            QtCore.Qt.ItemDataRole.BackgroundRole: epyqlib.nv.reference_highlight,
                        },
                    )
                    diff_proxy.setSourceModel(sort_proxy)

                    view.setModel(diff_proxy)
                    view.configure_sort_proxy(sort_proxy)
                    view.configure_diff_proxy(diff_proxy)

                    view.set_metas(self.metas)
                    view.set_sorting_enabled(True)
                    view.sort_by_column(column=column, order=Qt.AscendingOrder)

                    nv_range_check_overridable = self.raw_dict.get(
                        "nv_range_check_overridable",
                        False,
                    )

                    view.ui.enforce_range_limits_check_box.setVisible(
                        nv_range_check_overridable,
                    )

        if Elements.variables in self.elements:
            variable_model = epyqlib.variableselectionmodel.VariableModel(
                nvs=self.nvs,
                nv_model=nv_model,
                bus=self.bus,
                tx_id=self.neo_frames.frame_by_name("CCP").id,
                rx_id=self.neo_frames.frame_by_name("CCPResponse").id,
            )

            column = epyqlib.variableselectionmodel.Columns.indexes.name
            proxy = epyqlib.utils.qt.PySortFilterProxyModel(
                filter_column=column,
            )
            proxy.setSortCaseSensitivity(Qt.CaseInsensitive)
            proxy.setSourceModel(variable_model)
            self.ui.variable_selection.set_model(proxy)
            self.ui.variable_selection.set_sorting_enabled(True)
            self.ui.variable_selection.sort_by_column(
                column=column, order=Qt.AscendingOrder
            )
            self.ui.variable_selection.set_signal_paths(
                reset_signal_path=can_configuration.data_logger_reset_signal_path,
                recording_signal_path=can_configuration.data_logger_recording_signal_path,
                configuration_is_valid_signal_path=can_configuration.data_logger_configuration_is_valid_signal_path,
            )

        if Elements.fault_log in self.elements:
            fault_log = epyqlib.faultlogmodel.create_blank()

            nv_message_names = self.raw_dict["fault_log"]["nv_message_names"]
            process_message_names = self.raw_dict["fault_log"]["process_message_names"]

            fault_log.connect(
                process_frames=self.neo_frames,
                process_message_names=process_message_names,
                nv_frames=self.widget_nvs,
                nv_message_names=nv_message_names,
            )
            self.ui.fault_log_view.set_model(fault_log.model)

        if Tabs.dashes in tabs:
            for i, (name, dash) in enumerate(self.dash_uis.items()):
                self.ui.tabs.insertTab(i, dash, name)
        if Tabs.txrx not in tabs:
            self.ui.tabs.removeTab(self.ui.tabs.indexOf(self.ui.txrx))
        if Tabs.variables not in tabs:
            self.ui.tabs.removeTab(self.ui.tabs.indexOf(self.ui.variables))
        if Tabs.nv not in tabs:
            self.ui.tabs.removeTab(self.ui.tabs.indexOf(self.ui.nv))
        else:
            self.ui.tabs.currentChanged.connect(self.tab_changed)
        if Tabs.scripting not in tabs:
            self.ui.tabs.removeTab(self.ui.tabs.indexOf(self.ui.scripting))
        if Tabs.fault_log not in tabs:
            self.ui.tabs.removeTab(self.ui.tabs.indexOf(self.ui.faultlog))
        if Tabs.files not in tabs:
            self.ui.tabs.removeTab(self.ui.tabs.indexOf(self.ui.files))
        if Tabs.information not in tabs:
            self.ui.tabs.removeTab(self.ui.tabs.indexOf(self.ui.information))

        self.ui.tabs.setCurrentIndex(0)

        self.widget_nv_frames = collections.defaultdict(list)

        def flatten(dict_node):
            flat = set()
            for key, value in dict_node.items():
                if isinstance(value, dict):
                    flat |= flatten(value)
                else:
                    flat.add(value)

            return flat

        flat = flatten(self.dash_uis)
        flat = [v for v in flat if isinstance(v, QWidget)]

        default_widget_value = math.nan

        self.dash_connected_signals = set()
        self.dash_missing_signals = set()
        self.dash_missing_defaults = set()
        self.nv_looping_reads = {}
        if Tabs.variables in tabs:
            flat.append(self.ui.variable_selection)
        if Tabs.nv in tabs:
            flat.append(self.ui.nv)
        for dash in flat:
            # TODO: CAMPid 99457281212789437474299
            children = dash.findChildren(QObject)
            widgets = [c for c in children if isinstance(c, AbstractWidget)]

            dash.connected_frames = set()
            frames = dash.connected_frames

            for widget in widgets:
                # TODO: CAMPid 07340793413419714301373147
                widget.set_range(min=0, max=100)
                try:
                    widget.set_value(default_widget_value)
                except ValueError:
                    widget.set_value(0)

                frame = widget.property("frame")
                if frame is not None:
                    signal = widget.property("signal")
                    signal_path = (frame, signal)
                else:
                    signal_path = tuple(e for e in widget._signal_path if len(e) > 0)

                try:
                    signal = self.neo_frames.signal_by_path(*signal_path)
                except epyqlib.canneo.NotFoundError:
                    if not widget.ignore:
                        widget_path = []
                        p = widget
                        while p is not dash:
                            widget_path.insert(0, p.objectName())
                            p = p.parent()

                        self.dash_missing_signals.add(
                            "{}:/{} - {}".format(
                                (
                                    dash.file_name
                                    if hasattr(dash, "file_name")
                                    else "<builtin>"
                                ),
                                "/".join(widget_path),
                                ":".join(signal_path)
                                if len(signal_path) > 0
                                else "<none specified>",
                            )
                        )
                else:
                    # TODO: CAMPid 079320743340327834208
                    if signal.frame.id == self.nvs.set_frames[0].id:
                        nv_signal = self.widget_nvs.neo.signal_by_path(*signal_path)

                        self.widget_nv_frames[nv_signal.frame].append(
                            nv_signal,
                        )

                        if nv_signal.multiplex not in self.nv_looping_reads:

                            def read(
                                nv_signal=nv_signal,
                                read=self.nvs.protocol.read,
                            ):
                                d = read(
                                    nv_signal=nv_signal,
                                    meta=epyqlib.nv.MetaEnum.value,
                                )

                                d.addErrback(ignore_timeout)

                                return d

                            self.nv_looping_reads[nv_signal.multiplex] = read

                        if dash is self.ui.nv:
                            self.nv_tab_looping_set.add_request(
                                key=widget,
                                request=epyqlib.twisted.loopingset.Request(
                                    f=self.nv_looping_reads[nv_signal.multiplex],
                                    period=1,
                                ),
                            )
                        else:
                            self.nv_looping_set.add_request(
                                key=widget,
                                request=epyqlib.twisted.loopingset.Request(
                                    f=self.nv_looping_reads[nv_signal.multiplex],
                                    period=1,
                                ),
                            )

                        if hasattr(widget, "tx") and widget.tx:
                            signal = self.widget_nvs.neo.signal_by_path(
                                self.nvs.set_frames[0].name, *signal_path[1:]
                            )
                        else:
                            signal = self.widget_nvs.neo.signal_by_path(
                                self.nvs.status_frames[0].name, *signal_path[1:]
                            )

                    frame = signal.frame
                    frames.add(frame)
                    self.dash_connected_signals.add(signal)
                    widget.set_signal(signal)

                    if hasattr(widget, "tx") and widget.tx:
                        self.first_nv_view.ui.enforce_range_limits_check_box.stateChanged.connect(
                            widget.set_check_range,
                        )
                        widget.set_check_range(
                            self.first_nv_view.ui.enforce_range_limits_check_box.checkState(),
                        )

                if edit_actions is not None:
                    # TODO: CAMPid 97453289314763416967675427
                    if widget.property("editable"):
                        for action in edit_actions:
                            if action[1](widget):
                                action[0](dash=dash, widget=widget, signal=widget.edit)
                                break

        monitor_frames = epyqlib.canneo.Neo(
            matrix=matrix,
            node_id_adjust=self.node_id_adjust,
        )
        monitor_frame = monitor_frames.frame_by_name(
            can_configuration.monitor_frame,
        )

        self.ui.overlay_stack.layout().setStackingMode(
            QStackedLayout.StackingMode.StackAll,
        )
        self.ui.overlays.setAttribute(Qt.WA_TransparentForMouseEvents)

        self.connection_monitor = FrameTimeout(frame=monitor_frame)
        self.connection_monitor.lost.connect(self.connection_status_changed)
        self.connection_monitor.found.connect(self.connection_status_changed)
        self.connection_monitor.found.connect(self.read_nv_widget_min_max)
        self.connection_monitor.start()

        notifiees.append(self.connection_monitor)

        self.bus_status_changed(online=False, transmit=False, skip_set_stale=True)

        all_signals = set()
        for frame in self.neo_frames.frames:
            for signal in frame.signals:
                if signal.name != "__padding__":
                    all_signals.add(signal)

        frame_signals = []
        for signal in all_signals - self.dash_connected_signals:
            frame_signals.append("{} : {}".format(signal.frame.name, signal.name))

        if Elements.nv in self.elements:
            nv_frame_signals = []
            for frame in list(self.nvs.set_frames.values()) + list(
                self.nvs.status_frames.values()
            ):
                for signal in frame.signals:
                    nv_frame_signals.append(
                        "{} : {}".format(signal.frame.name, signal.name)
                    )

            frame_signals = list(set(frame_signals) - set(nv_frame_signals))

        if len(frame_signals) > 0:
            logger.warning("\n === Signals not referenced by a widget")
            for frame_signal in sorted(frame_signals):
                logger.warning(frame_signal)

        if len(self.dash_missing_signals) > 0:
            logger.error("\n === Signals referenced by a widget but not defined")
            undefined_signals = sorted(self.dash_missing_signals)
            logger.error("\n".join(undefined_signals))

            message = (
                "The following signals are referenced by the .ui "
                "files but were not found in the loaded CAN "
                "database.  The widgets will show `{}`.".format(default_widget_value)
            )

            message = textwrap.dedent(
                """\
            {message}

            {signals}
            """
            ).format(message=message, signals="\n\n".join(undefined_signals))

            epyqlib.utils.qt.dialog(
                parent=None,
                message=message,
                icon=QMessageBox.Information,
            )

        if Elements.scripting in self.elements:
            scripting_model = epyqlib.scripting.Model(
                get_devices=lambda: {None: self, self.nickname: self},
            )
            self.ui.scripting_view.set_model(scripting_model)

        self.device_interface = DeviceInterface(device=self)

        if Tabs.files in tabs:
            self.ui.files_view.set_device_interface(self.device_interface)

        if Tabs.information in tabs:
            self.ui.information_text_browser.append(f"Name:\t\t{self.name}")
            self.ui.information_text_browser.append(f"Type:\t\t{self.node_id_type}")
            self.ui.information_text_browser.append(f"SHA:\t\t{self.device_sha}")
            self.ui.information_text_browser.append(f"File:\t\t{self.device_file}")

        self.notifiees = notifiees
        for notifiee in notifiees:
            self.bus.notifier.add(notifiee)

        self.extension.post()

    def tab_changed(self, index):
        tabs = {self.ui.tabs.indexOf(x) for x in (self.ui.nv, self.ui.scripting)}
        if index in tabs:
            self.nv_looping_set.stop()
            self.nv_tab_looping_set.start()
        else:
            self.nv_looping_set.start()
            self.nv_tab_looping_set.stop()

        # TODO: Remove this and find a better way to manage when files view should fetch its list
        if self.ui.tabs.indexOf(self.ui.files) == index:
            self.ui.files_view.tab_selected()

    def absolute_path(self, path=""):
        # TODO: CAMPid 9549757292917394095482739548437597676742
        if not QFileInfo(path).isAbsolute():
            path = os.path.join(
                QFileInfo.absolutePath(QFileInfo(self.config_path)), path
            )

        return path

    def get_frames(self):
        return self.frames

    @pyqtSlot(bool, bool, bool)
    def bus_status_changed(self, online, transmit, skip_set_stale=False):
        self.bus_online = online
        self.bus_tx = transmit

        style = epyqlib.overlaylabel.styles["red"]
        text = ""
        if online:
            if not transmit:
                text = "passive"
                style = epyqlib.overlaylabel.styles["blue"]
        else:
            text = "offline"

        self.ui.offline_overlay.ui.label.setText(text)
        self.ui.offline_overlay.setVisible(len(text) > 0)
        self.ui.offline_overlay.setStyleSheet(style)

        self.read_nv_widget_min_max()

        if not skip_set_stale:
            if self.nvs is not None:
                if not online:
                    self.nvs.set_stale()

        if self.ui.tabs.indexOf(self.ui.files) != -1:
            self.ui.files_view.on_bus_status_changed()

    def read_nv_widget_min_max(self):
        if not self.auto_read_nv_widget_min_max:
            return

        logger.info("bus_online", self.bus_online)
        logger.info("bus_tx", self.bus_tx)
        logger.info("present", self.connection_monitor.present)
        active = all(
            (
                self.bus_online,
                self.bus_tx,
                self.connection_monitor.present,
            )
        )

        if not active:
            return

        logger.info("reading min/max for nv widgets")

        metas = (
            epyqlib.nv.MetaEnum.minimum,
            epyqlib.nv.MetaEnum.maximum,
        )

        for frame, signals in self.widget_nv_frames.items():
            for meta in metas:
                d = self.nvs.protocol.read_multiple(
                    nv_signals=signals,
                    meta=meta,
                )

                d.addErrback(ignore_timeout)

    def connection_status_changed(self):
        present = self.connection_monitor.present

        text = "no status"
        if present:
            text = ""

        self.ui.connection_monitor_overlay.ui.label.setText(text)
        self.ui.connection_monitor_overlay.setVisible(len(text) > 0)

        if self.nvs is not None:
            if present:
                self.nvs.cyclic_read_all()
            else:
                self.nvs.set_stale()
                self.nvs.cancel_cyclic_read_all()


class FrameTimeout(epyqlib.canneo.QtCanListener):
    lost = epyqlib.utils.qt.Signal()
    found = epyqlib.utils.qt.Signal()

    def __init__(self, frame, relative=lambda t: 5 * t, absolute=0.5, parent=None):
        super().__init__(self.message_received, parent=parent)

        self.frame = frame

        self.timeout = max(
            absolute,
            relative(float(self.frame.cycle_time) / 1000),
        )

        self.present = False

        self.timer = QtCore.QTimer()
        self.timer.timeout.connect(self._lost)

    def _lost(self):
        self.timer.stop()
        self.present = False
        self.lost.emit()

    def start(self):
        self._lost()

    def message_received(self, msg):
        if not self.frame.message_received(msg):
            return

        self.timer.start(1000 * self.timeout)

        if not self.present:
            self.present = True
            self.found.emit()

    def terminate(self):
        self.timer.stop()


@attr.s
class DeviceInterface:
    device: Device = attr.ib()

    @attr.s(slots=True, auto_attribs=True)
    class TransmitStatus:
        connected: bool
        transmitting: bool

    async def get_serial_number(self) -> str:
        serial_number_signal = self.device.nvs.signal_from_names(
            "SN",
            "SerialNumber",
        )

        return await self._read_single_param(serial_number_signal)

    def get_serial_number_from_parameters(self) -> str:
        serial_number_signal = self.device.nvs.signal_from_names(
            "SN",
            "SerialNumber",
        )

        return str(serial_number_signal.scaled_value)

    async def get_build_hash(self) -> str:
        serial_number_signal = self.device.nvs.signal_from_names(
            "SoftwareHash",
            "SoftwareHash",
        )

        return await self._read_single_param(serial_number_signal)

    async def _read_single_param(self, param_signal):
        param_response = await self.device.nvs.read_all_from_device(
            only_these=(param_signal,),
            meta=(epyqlib.nv.MetaEnum.value,),
        )
        param = str(
            param_response[0][param_signal.status_signal],
        )
        return param

    def get_connected_status(self) -> TransmitStatus:
        return DeviceInterface.TransmitStatus(
            connected=self.device.bus_online, transmitting=self.device.bus_tx
        )

    def on_offline_bus(self):
        return self.device.on_offline_bus


if __name__ == "__main__":
    import sys

    logger.info("No script functionality here")
    sys.exit(1)  # non-zero is a failure



import os
import textwrap

import click


@click.command()
@click.argument("target", type=click.File("w"), default="-")
def write_build_file(target):
    template = textwrap.dedent(
        """\
    # This file has been generated


    """
    )

    target.write(template)

    values = {
        name: None
        for name in (
            "build_system",
            "build_id",
            "build_number",
            "build_version",
            "job_id",
            "job_url",
        )
    }

    if os.environ.get("APPVEYOR") == "True":
        values["build_system"] = "AppVeyor"
        mapping = {
            "build_id": "APPVEYOR_BUILD_ID",
            "build_number": "APPVEYOR_BUILD_NUMBER",
            "build_version": "APPVEYOR_BUILD_VERSION",
            "job_id": "APPVEYOR_JOB_ID",
        }
        values.update({k: os.environ[v] for k, v in mapping.items()})
        values[
            "job_url"
        ] = "https://ci.appveyor.com/" "project/{account}/{slug}/build/job/{id}".format(
            account=os.environ["APPVEYOR_ACCOUNT_NAME"],
            slug=os.environ["APPVEYOR_PROJECT_SLUG"],
            id=os.environ["APPVEYOR_JOB_ID"],
        )
    elif os.environ.get("GITHUB_ACTIONS") == "true":
        GITHUB_RUN_ID = os.environ.get("GITHUB_RUN_ID", "None")
        GITHUB_RUN_NUMBER = os.environ.get("GITHUB_RUN_NUMBER", "None")
        GITHUB_SERVER_URL = os.environ.get("GITHUB_SERVER_URL", "None")
        GITHUB_REPOSITORY = os.environ.get("GITHUB_REPOSITORY", "None")
        values["build_system"] = "GitHub Actions"
        values["job_id"] = f"{GITHUB_RUN_ID}_{GITHUB_RUN_NUMBER}"
        mapping = {
            "build_id": "GITHUB_RUN_ID",
            "build_number": "GITHUB_RUN_NUMBER",
            # "build_version": "APPVEYOR_BUILD_VERSION",
            # "job_id": "GITHUB_JOB",
        }
        values.update({k: os.environ[v] for k, v in mapping.items()})
        values[
            "job_url"
        ] = f"{GITHUB_SERVER_URL}/{GITHUB_REPOSITORY}/actions/runs/{GITHUB_RUN_ID}"

    for k, v in values.items():
        target.write("{} = {}\n".format(k, repr(v)))
